# ---
# jupyter:
#   jupytext:
#     cell_metadata_filter: -all
#     custom_cell_magics: kql
#     formats: ipynb,py:percent
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.19.1
#   kernelspec:
#     display_name: .venv
#     language: python
#     name: python3
# ---

# %% [markdown]
# # Spatial Modelling with Composable Gaussian Processes
#
# This notebook shows how to construct a semiparametric linear model by composing a linear model
# in NumPyro with a GPJax Gaussian Process (GP).
# We build two components: firstly, the linear component which ncodes a global affine trend with
# Bayesian linear regression. We then define a GP residual component which is responsible for
# capturing spatial structure that the linear term's residual.
#
# The example highlights the interplay between **GPJax** and **NumPyro**: `GPJax` provides the GP
# prior and likelihood definitions, while `NumPyro` performs Hamiltonian Monte Carlo (HMC)
# inference across all parameters in a unified model and allows us to draw upon a broader set of
# modelling components.

# %% [markdown]
# ## Data Simulation
#
# We simulate a 2D spatial dataset ($N=200$) on a domain $[0, 5] \times [0, 5]$. The generative
# process contains a linear trend: $y_{\text{lin}} = 2x_1 - 1x_2 + 1.5$ with an additive spatial
# residual: $y_{\text{res}} = \sin(x_1) \cos(x_2)$. To this, we add simulate an additive
# homoscedastic noise component $\epsilon \sim \mathcal{N}(0, 0.1^2)$. The dominant linear trend
# masks a non-linear residual. Composing models lets us represent both
# behaviours without forcing a single mechanism to fit every feature of the data.

# %%
from functools import partial

import jax
import jax.numpy as jnp
import jax.random as jr
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpyro
import numpyro.distributions as dist
from numpyro.infer import (
    MCMC,
    NUTS,
    Predictive,
)

from examples.utils import use_mpl_style
import gpjax as gpx
from gpjax.numpyro_extras import register_parameters

jax.config.update("jax_enable_x64", True)

use_mpl_style()
cols = mpl.rcParams["axes.prop_cycle"].by_key()["color"]


N = 200
key = jr.key(123)
keys = jr.split(key, 8)

X = jr.uniform(keys[0], shape=(N, 2), minval=0.0, maxval=5.0)

# True Linear Trend
true_slope = jnp.array([2.0, -1.0])
true_intercept = 1.5
y_lin = X @ true_slope + true_intercept

# Non-linear Spatial Residual
y_res = jnp.sin(X[:, 0]) * jnp.cos(X[:, 1])

# Total Signal + Noise
latent_signal = y_lin + y_res
noise_stddev = 0.1
y = latent_signal + noise_stddev * jr.normal(keys[1], shape=latent_signal.shape)

# %% [markdown]
# ## Linear Component
#
# We begin by defining a Bayesian linear regression model in NumPyro. This component
# will later be combined with a GP residual, but for now, we'll establish a baseline model
# through ordinary linear regression.
#
# $$\begin{aligned} \mathbf{w} &\sim \mathcal{N}(\mathbf{0}, 5\mathbf{I}) \\
# b &\sim \mathcal{N}(0, 5) \\
# \sigma &\sim \text{LogNormal}(0, 1) \\
# \mathbf{y} &\sim \mathcal{N}(\mathbf{X}\mathbf{w} + b, \sigma^2 \mathbf{I}) \end{aligned} $$
#
# We use the No-U-Turn Sampler (NUTS) to draw samples from the posterior distributions of the slope
# $\mathbf{w}$, intercept $b$, and noise $\sigma$.


# %%
def linear_model(X, Y=None):
    slope = numpyro.sample("slope", dist.Normal(0.0, 5.0).expand([2]))
    intercept = numpyro.sample("intercept", dist.Normal(0.0, 5.0))
    obs_noise = numpyro.sample("obs_noise", dist.LogNormal(0.0, 1.0))

    mu = X @ slope + intercept
    numpyro.deterministic("mu", mu)
    numpyro.sample("obs", dist.Normal(mu, obs_noise), obs=Y)


nuts_kernel_lin = NUTS(linear_model)
mcmc_lin = MCMC(nuts_kernel_lin, num_warmup=1500, num_samples=2000, num_chains=1)
mcmc_lin.run(keys[2], X, y)
mcmc_lin.print_summary()

# %% [markdown]
# ## Composing the Linear Component with a GP
#
# We now augment the linear component with a GP tasked with modelling the residual.
#
# $$ y(\mathbf{x}) = \underbrace{\mathbf{w}^T \mathbf{x} + b}_{\text{Linear Mean}} +
# \underbrace{f(\mathbf{x})}_{\text{GP Residual}} + \epsilon $$
#
# ### GPJax and NumPyro Integration
#
# We define the GP prior in `GPJax` using an second-order Matérn kernel and a constant mean function
# (since the linear trend is handled explicitly). We attach `dist.LogNormal` priors to
# the kernel's hyperparameters (lengthscale and variance) directly within the GPJax object.
# We then register the parameters by calling
# `gpx.numpyro_extras.register_parameters(gp_posterior)` inside the NumPyro model. This
# function traverses the GPJax object, identifies parameters with attached priors, and
# registers them as NumPyro sample sites. It returns a new GPJax object where the parameters
# have been replaced by the values sampled by NumPyro. Finally, we compute the exact marginal
# log-likelihood (MLL) of the residuals under the GP prior using `gpx.objectives.conjugate_mll`.
# This term is added to the potential function using `numpyro.factor`, guiding the sampler.

# %%
lengthscale = gpx.parameters.PositiveReal(1.0, prior=dist.LogNormal(0.0, 1.0))
variance = gpx.parameters.PositiveReal(1.0, prior=dist.LogNormal(0.0, 1.0))

kernel = gpx.kernels.Matern32(
    active_dims=[0, 1], lengthscale=lengthscale, variance=variance
)
meanf = gpx.mean_functions.Constant()
prior = gpx.gps.Prior(mean_function=meanf, kernel=kernel)

obs_stddev = gpx.parameters.NonNegativeReal(0.1, prior=dist.LogNormal(0.0, 1.0))
likelihood = gpx.likelihoods.Gaussian(num_datapoints=N, obs_stddev=obs_stddev)
gp_posterior = prior * likelihood


def joint_model(X, Y, gp_posterior, X_new=None):
    slope = numpyro.sample("slope", dist.Normal(0.0, 5.0).expand([2]))
    intercept = numpyro.sample("intercept", dist.Normal(0.0, 5.0))

    trend = X @ slope + intercept

    p_posterior = register_parameters(gp_posterior)

    if Y is not None:
        residuals = Y - trend
        residuals = residuals.reshape(-1, 1)
        D_resid = gpx.Dataset(X=X, y=residuals)
        mll = gpx.objectives.conjugate_mll(p_posterior, D_resid)
        numpyro.factor("gp_log_lik", mll)

    if X_new is not None:
        if Y is not None:
            residuals = Y - trend
            residuals = residuals.reshape(-1, 1)
            D_resid = gpx.Dataset(X=X, y=residuals)

            latent_dist = p_posterior.predict(X_new, train_data=D_resid)
            f_new = numpyro.sample("f_new", latent_dist)
            f_new = f_new.reshape((-1, 1))

            total_prediction = (X_new @ slope + intercept).reshape(-1, 1) + f_new
            numpyro.deterministic("y_pred", total_prediction)


joint_model_wrapper = partial(joint_model, gp_posterior=gp_posterior)
nuts_kernel_joint = NUTS(joint_model_wrapper)
# In practice, one should run more samples from multiple chains.
mcmc_joint = MCMC(nuts_kernel_joint, num_warmup=1500, num_samples=2000, num_chains=1)
mcmc_joint.run(keys[3], X, y)
mcmc_joint.print_summary()

# %% [markdown]
# ## Comparison and Visualization
#
# We evaluate the linear model in isolation, and then the joint model where a GP has
# been included to model the residual.

# %%
samples_lin = mcmc_lin.get_samples()
predictive_lin = Predictive(linear_model, samples_lin, return_sites=["mu"])
preds_lin = predictive_lin(keys[4], X=X)["mu"]
mean_pred_lin = jnp.mean(preds_lin, axis=0)

samples_joint = mcmc_joint.get_samples()
predictive_joint = Predictive(
    joint_model_wrapper, samples_joint, return_sites=["y_pred"]
)
preds_joint = predictive_joint(keys[5], X=X, Y=y, X_new=X)["y_pred"]
mean_pred_joint = jnp.mean(preds_joint, axis=0)

rmse_lin = jnp.sqrt(jnp.mean((mean_pred_lin.flatten() - latent_signal.flatten()) ** 2))
rmse_joint = jnp.sqrt(
    jnp.mean((mean_pred_joint.flatten() - latent_signal.flatten()) ** 2)
)

print("\nRMSE Comparison (vs True Signal):")
print(f"Linear Model: {rmse_lin:.4f}")
print(f"Joint Model:  {rmse_joint:.4f}")

# %% [markdown]
# Let's now plot the predicted profiles from both models.

# %%
n_grid = 30
x1 = jnp.linspace(0, 5, n_grid)
x2 = jnp.linspace(0, 5, n_grid)
X1, X2 = jnp.meshgrid(x1, x2)
X_grid = jnp.column_stack([X1.ravel(), X2.ravel()])

y_grid_true = (X_grid @ true_slope + true_intercept) + (
    jnp.sin(X_grid[:, 0]) * jnp.cos(X_grid[:, 1])
)

preds_lin_grid = predictive_lin(keys[6], X=X_grid)["mu"]
mean_pred_lin_grid = jnp.mean(preds_lin_grid, axis=0)

preds_joint_grid = predictive_joint(keys[7], X=X, Y=y, X_new=X_grid)["y_pred"]
mean_pred_joint_grid = jnp.mean(preds_joint_grid, axis=0)

fig, axes = plt.subplots(1, 3, figsize=(12, 3), sharey=True)

vmin = min(y_grid_true.min(), mean_pred_lin_grid.min(), mean_pred_joint_grid.min())
vmax = max(y_grid_true.max(), mean_pred_lin_grid.max(), mean_pred_joint_grid.max())
levels = jnp.linspace(vmin, vmax, 20)

c0 = axes[0].tricontourf(
    X_grid[:, 0], X_grid[:, 1], y_grid_true, levels=levels, cmap="magma"
)
axes[0].set_title("True Signal")

c1 = axes[1].tricontourf(
    X_grid[:, 0],
    X_grid[:, 1],
    mean_pred_lin_grid.flatten(),
    levels=levels,
    cmap="magma",
)
axes[1].set_title(f"Linear Model (RMSE: {rmse_lin:.2f})")

c2 = axes[2].tricontourf(
    X_grid[:, 0],
    X_grid[:, 1],
    mean_pred_joint_grid.flatten(),
    levels=levels,
    cmap="magma",
)
axes[2].set_title(f"Joint Model (RMSE: {rmse_joint:.2f})")

cbar = fig.colorbar(c0, ax=axes.tolist())
cbar.ax.yaxis.set_major_formatter(mpl.ticker.FormatStrFormatter("%d"))

for ax in axes:
    ax.set_xlabel("x1")
    ax.scatter(X[:, 0], X[:, 1], c=cols[0], s=10, alpha=0.5)


# %% [markdown]
# ## System configuration

# %%
# %load_ext watermark
# %watermark -n -u -v -iv -w -a "Thomas Pinder"



from gpjax.kernels.approximations import RFF
from gpjax.kernels.nonstationary import (
    Linear,
    Polynomial,
)
from gpjax.kernels.stationary import (
    RBF,
    Matern12,
    Matern32,
    Matern52,
    Periodic,
    PoweredExponential,
    RationalQuadratic,
    StationaryKernel,
)
from gpjax.linalg.operators import Dense
import jax
from jax import config
import jax.numpy as jnp
import jax.random as jr
import pytest

config.update("jax_enable_x64", True)
_jitter = 1e-6


@pytest.mark.parametrize("kernel", [RBF, Matern12, Matern32, Matern52])
@pytest.mark.parametrize("num_basis_fns", [2, 10])
@pytest.mark.parametrize("n_dims", [1, 3])
def test_frequency_sampler(
    kernel: type[StationaryKernel], num_basis_fns: int, n_dims: int
):
    base_kernel = kernel(active_dims=list(range(n_dims)))
    approximate = RFF(base_kernel=base_kernel, num_basis_fns=num_basis_fns)
    assert approximate.frequencies.shape == (num_basis_fns, n_dims)


@pytest.mark.parametrize("kernel", [RBF, Matern12, Matern32, Matern52])
@pytest.mark.parametrize("num_basis_fns", [2, 10])
@pytest.mark.parametrize("n_dims", [1, 3])
@pytest.mark.parametrize("n_data", [50, 100])
def test_gram(
    kernel: type[StationaryKernel], num_basis_fns: int, n_dims: int, n_data: int
):
    key = jr.key(123)
    x = jr.uniform(key, shape=(n_data, 1), minval=-3.0, maxval=3.0).reshape(-1, 1)
    if n_dims > 1:
        x = jnp.hstack([x] * n_dims)
    base_kernel = kernel(active_dims=list(range(n_dims)))
    approximate = RFF(base_kernel=base_kernel, num_basis_fns=num_basis_fns)

    linop = approximate.gram(x)

    # Check the return type
    assert isinstance(linop, Dense)

    Kxx = linop.to_dense() + jnp.eye(n_data) * _jitter

    # Check that the shape is correct
    assert Kxx.shape == (n_data, n_data)

    # Check that the Gram matrix is PSD
    evals, _ = jnp.linalg.eigh(Kxx)
    assert jnp.all(evals > 0)


@pytest.mark.parametrize("kernel", [RBF, Matern12, Matern32, Matern52])
@pytest.mark.parametrize("num_basis_fns", [2, 10])
@pytest.mark.parametrize("n_dims", [1, 3])
@pytest.mark.parametrize("n_datas", [(50, 100), (100, 50)])
def test_cross_covariance(
    kernel: type[StationaryKernel],
    num_basis_fns: int,
    n_dims: int,
    n_datas: tuple[int, int],
):
    nd1, nd2 = n_datas
    key = jr.key(123)
    x1 = jr.uniform(key, shape=(nd1, 1), minval=-3.0, maxval=3.0)
    if n_dims > 1:
        x1 = jnp.hstack([x1] * n_dims)
    x2 = jr.uniform(key, shape=(nd2, 1), minval=-3.0, maxval=3.0)
    if n_dims > 1:
        x2 = jnp.hstack([x2] * n_dims)

    base_kernel = kernel(active_dims=list(range(n_dims)))
    approximate = RFF(base_kernel=base_kernel, num_basis_fns=num_basis_fns)
    Kxx = approximate.cross_covariance(x1, x2)

    # Check the return type
    assert isinstance(Kxx, jax.Array)

    # Check that the shape is correct
    assert Kxx.shape == (nd1, nd2)


@pytest.mark.parametrize("kernel", [RBF, Matern12, Matern32, Matern52])
@pytest.mark.parametrize("n_dim", [1, 3])
def test_improvement(kernel: type[StationaryKernel], n_dim: int):
    n_data = 100
    key = jr.key(123)

    x = jr.uniform(key, minval=-3.0, maxval=3.0, shape=(n_data, n_dim))
    base_kernel = kernel(active_dims=list(range(n_dim)))
    exact_linop = base_kernel.gram(x).to_dense()

    crude_approximation = RFF(base_kernel=base_kernel, num_basis_fns=10)
    c_linop = crude_approximation.gram(x).to_dense()

    better_approximation = RFF(base_kernel=base_kernel, num_basis_fns=100)
    b_linop = better_approximation.gram(x).to_dense()

    c_delta = jnp.linalg.norm(exact_linop - c_linop, ord="fro")
    b_delta = jnp.linalg.norm(exact_linop - b_linop, ord="fro")

    # The frobenius norm of the difference between the exact and approximate
    # should improve as we increase the number of basis functions
    assert c_delta > b_delta


@pytest.mark.parametrize("kernel", [RBF, Matern12, Matern32, Matern52])
def test_exactness(kernel: type[StationaryKernel]):
    kernel = kernel(n_dims=1)

    n_data = 100
    key = jr.key(123)

    x = jr.uniform(key, minval=-3.0, maxval=3.0, shape=(n_data, 1))
    exact_linop = kernel.gram(x).to_dense()

    better_approximation = RFF(base_kernel=kernel, num_basis_fns=300)
    b_linop = better_approximation.gram(x).to_dense()

    max_delta = jnp.max(exact_linop - b_linop)
    assert max_delta < 0.1


@pytest.mark.parametrize("kernel", [Polynomial, Linear])
def test_nonstationary_raises_error(kernel):
    with pytest.raises(TypeError):
        RFF(base_kernel=kernel(1), num_basis_fns=10)


@pytest.mark.parametrize(
    "kernel",
    [RationalQuadratic, PoweredExponential, Periodic],
)
def test_missing_spectral_density_raises_error(kernel):
    with pytest.raises(NotImplementedError):
        RFF(base_kernel=kernel(), num_basis_fns=10)


@pytest.mark.parametrize("kernel", [RBF, Matern12, Matern32, Matern52])
def test_stochastic_init(kernel: type[StationaryKernel]):
    with pytest.raises(ValueError):
        # n_dims is not specified, but should be
        RFF(base_kernel=kernel(), num_basis_fns=10)

    kernel = kernel(n_dims=1)

    k1 = RFF(base_kernel=kernel, num_basis_fns=10, key=jr.key(123))
    k2 = RFF(base_kernel=kernel, num_basis_fns=10, key=jr.key(42))

    assert (k1.frequencies != k2.frequencies).any()



# ---
# jupyter:
#   jupytext:
#     cell_metadata_filter: -all
#     custom_cell_magics: kql
#     formats: ipynb,py:percent
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.19.1
#   kernelspec:
#     display_name: python3
#     language: python
#     name: python3
# ---

# %% [markdown]
# # Joint Inference with Numpyro
#
# In this notebook, we demonstrate how to use [Numpyro](https://num.pyro.ai/) to perform fully
# Bayesian inference over the hyperparameters of a Gaussian process model.  We will look at a
# scenario where we have a structured mean function in the form of a linear model, and a GP
# capturing the residuals. We will infer the parameters of both the linear model and the GP jointly.

# %%
from jax import config
import jax.numpy as jnp
import jax.random as jr
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpyro
import numpyro.distributions as dist
from numpyro.infer import (
    MCMC,
    NUTS,
    Predictive,
)

from examples.utils import use_mpl_style
import gpjax as gpx
from gpjax.numpyro_extras import register_parameters

config.update("jax_enable_x64", True)

use_mpl_style()
cols = mpl.rcParams["axes.prop_cycle"].by_key()["color"]

key = jr.key(123)
keys = jr.split(key, 4)

# %% [markdown]
# ## Data Generation
#
# We generate a synthetic dataset that consists of a linear trend together with a locally periodic
# residual signal whose amplitude varies over time, an additional high-frequency component, and a
# local bump. This data generating process is purposefully designed to illustrate the benefit of
# incorporating a Gaussian process into a larger Bayesian model; however, such structures are
# common.

# %%
N = 200

x = jnp.sort(jr.uniform(keys[0], shape=(N, 1), minval=0.0, maxval=10.0), axis=0)

# True parameters for the linear trend
true_slope = 0.45
true_intercept = 1.5

# Structured residual signal captured by the GP
slow_period = 6.0
fast_period = 0.8
amplitude_envelope = 1.0 + 0.5 * jnp.sin(2 * jnp.pi * x / slow_period)
modulated_periodic = amplitude_envelope * jnp.sin(2 * jnp.pi * x / fast_period)
high_frequency_component = 0.3 * jnp.cos(2 * jnp.pi * x / 0.35)
localised_bump = 1.2 * jnp.exp(-0.5 * ((x - 7.0) / 0.45) ** 2)

linear_trend = true_slope * x + true_intercept
residual_signal = modulated_periodic + high_frequency_component + localised_bump
signal = linear_trend + residual_signal

# Observations with homoscedastic noise
observation_noise = 0.7
y = signal + observation_noise * jr.normal(keys[1], shape=x.shape)

D = gpx.Dataset(X=x, y=y)

fig, ax = plt.subplots()
ax.plot(x, y, "o", label="Observations", color=cols[0])
ax.plot(x, signal, "--", label="True Signal", color=cols[1])
ax.legend()

# %% [markdown]
# ## Model Definition
#
# We define a GP model with a zero mean function, as we will handle the linear
# trend explicitly in the Numpyro model. Naturally, one could parameterise the GP with a
# linear mean function; however, this design is purely pedagogical. For the kernel, we specify a
# product of a periodic kernel and an RBF kernel. This choice reflects our prior knowledge that
# the signal is locally periodic. For a more in-depth look at how complex kernels can be designed,
# see our
# [Introduction to Kernels](https://docs.jaxgaussianprocesses.com/_examples/intro_to_kernels/)
# notebook.
#
# We see in the below that priors are specified on the parameters' constrained space. For
# example, the lengthscale parameter must be strictly positive and, therefore, a unit-Gaussian
# would be a poor choice of prior. Instead, we opt for the log-Gaussian as the prior distribution
# as its support matches that of our lengthscale parameter. Attaching a prior to a parameter is
# straightforward using the `prior` argument in the parameter's class and specifying any
# [numpyro distribution](https://num.pyro.ai/en/stable/distributions.html).

# %%
# Define priors
lengthscale_prior = dist.LogNormal(0.0, 1.0)
variance_prior = dist.LogNormal(0.0, 1.0)
period_prior = dist.LogNormal(0.0, 0.5)
noise_prior = dist.LogNormal(0.0, 1.0)

# We can explicitly attach priors to the parameters
lengthscale = gpx.parameters.PositiveReal(1.0, prior=lengthscale_prior)
variance = gpx.parameters.PositiveReal(1.0, prior=variance_prior)
period = gpx.parameters.PositiveReal(1.0, prior=period_prior)
noise = gpx.parameters.NonNegativeReal(1.0, prior=noise_prior)

# %% [markdown]
#
# Now that all of our parameters are defined, we'll proceed to construct the Gaussian process in
# the ordinary fashion. For a deeper look at how this is done, our
# [Regression](https://docs.jaxgaussianprocesses.com/_examples/regression/)
# notebook is a good starting point.

# %%
stationary_component = gpx.kernels.RBF(
    lengthscale=lengthscale,
    variance=variance,
)
periodic_component = gpx.kernels.Periodic(
    lengthscale=lengthscale,
    period=period,
)
kernel = stationary_component * periodic_component

meanf = gpx.mean_functions.Constant()
prior = gpx.gps.Prior(mean_function=meanf, kernel=kernel)

likelihood = gpx.likelihoods.Gaussian(
    num_datapoints=N,
    obs_stddev=noise,
)
posterior = prior * likelihood

# %% [markdown]
# ## Joint Inference Loop
#
# With a GPJax Posterior object now defined, the only outstanding task is to integrate
# it into a full Numpyro model. This notebook is not designed to be a full introduction to
# Numpyro (for that, see the excellent
# [Numpyro Documentation](https://num.pyro.ai/en/stable/)); however, in the below
# model we first sample the slope and intercept parameters of the linear component.
# We then compute the residuals between the observed data and the linear component,
# before then computing the GP marginal log-likelihood of the residual.
#
# The key step in the below is registering the parameters of the GPJax model with
# Numpyro via GPJax's `register_parameters` function. This function automatically
# samples parameters of the model and returns an updated state of the model with those
# sampled values used as parameters.


# %%
def model(X, Y, X_new=None):
    # 1. Sample linear model parameters
    slope = numpyro.sample("slope", dist.Normal(0.0, 2.0))
    intercept = numpyro.sample("intercept", dist.Normal(0.0, 2.0))
    linear_component = slope * X + intercept

    residuals = Y - linear_component

    p_posterior = register_parameters(posterior)
    D_resid = gpx.Dataset(X=X, y=residuals)
    mll = gpx.objectives.conjugate_mll(p_posterior, D_resid)

    numpyro.factor("gp_log_lik", mll)

    if X_new is not None:
        latent_dist = p_posterior.predict(X_new, train_data=D_resid)
        f_new = numpyro.sample("f_new", latent_dist)
        f_new = f_new.reshape((-1, 1))

        # Add observation noise to get noisy predictions
        obs_stddev = p_posterior.likelihood.obs_stddev.value
        y_noise = numpyro.sample(
            "y_noise",
            dist.Normal(0.0, obs_stddev).expand(f_new.shape).to_event(f_new.ndim),
        )

        total_prediction = slope * X_new + intercept + f_new + y_noise
        numpyro.deterministic("y_pred", total_prediction)
        return total_prediction


# %% [markdown]
# ## Running MCMC
#
# Using Numpyro's NUTS sampler, we can now draw samples from the posterior. To ensure
# our documentation can be quickly built, we limit the number of samples and the length
# of the burn-in phase below. However, in practice, one should draw more samples from
# multiple chains using the `num_chains` argument in the `MCMC` constructor.

# %%
nuts_kernel = NUTS(model)
# In practice, one should run more samples from multiple chains.
mcmc = MCMC(
    nuts_kernel,
    num_warmup=500,
    num_samples=500,
)
mcmc.run(keys[2], x, y)
mcmc.print_summary()

# %% [markdown]
# ## Analysis and Plotting
#
# Having obtained samples from the posterior, we now evaluate the predictive posterior
# at the test sites. In our
# [Poisson Regression](https://docs.jaxgaussianprocesses.com/_examples/poisson/), this
# process is done manually. However, by virtue of using Numpyro here, we may instead
# use Numpyro's `Predictive` object to handle this process for us. Once samples are
# drawn from the predictive posterior distribution, we may evaluate the mean and 95%
# credible interval and compare our model's predictions to the underlying data.

# %%
samples = mcmc.get_samples()
predictive = Predictive(
    model,
    posterior_samples=samples,
    return_sites=["y_pred"],
)

x_test = jnp.linspace(-0.5, 10.5, 1000).reshape(-1, 1)
predictions = predictive(keys[3], x, y, X_new=x_test)
y_pred = predictions["y_pred"]

mean_prediction = jnp.mean(y_pred, axis=0)
lower, upper = jnp.percentile(y_pred, jnp.array([2.5, 97.5]), axis=0)

fig, ax = plt.subplots()
ax.scatter(x, y, alpha=0.5, label="Observations", color=cols[0])
ax.plot(x, signal, "--", label="True Signal", color=cols[0])

ax.plot(x_test, mean_prediction, "-", label="Posterior Mean", color=cols[1])
ax.fill_between(
    x_test.flatten(),
    lower.flatten(),
    upper.flatten(),
    color=cols[1],
    alpha=0.2,
    label="95% Credible Interval",
)
ax.legend()

# %% [markdown]
# ## Conclusions
# This concludes our introduction to the integration of GPJax with Numpyro. The
# presentation given here is designed to best illustrate *how* the two libraries
# integrate. For a closer look at the more complex models that one may build by
# integrating Numpyro and GPJax, see our
# [Spatial Semi-Linear Model](https://docs.jaxgaussianprocesses.com/_examples/spatial_linear_gp)
# notebook.


# %% [markdown]
# ## System configuration

# %%
# %load_ext watermark
# %watermark -n -u -v -iv -w -a "Thomas Pinder"



from flax import nnx
from gpjax.numpyro_extras import (
    register_parameters,
    resolve_prior,
    tree_path_to_name,
)
from gpjax.parameters import (
    LowerTriangular,
    NonNegativeReal,
    PositiveReal,
    Real,
    SigmoidBounded,
)
from hypothesis import (
    given,
    strategies as st,
)
from hypothesis.extra.numpy import arrays
import jax.numpy as jnp
import jax.tree_util as jtu
import numpy as np
import numpyro.distributions as dist
from numpyro.handlers import (
    seed,
    trace,
)


def valid_shapes(min_dims=0, max_dims=2):
    return st.integers(min_dims, max_dims).flatmap(
        lambda d: st.lists(st.integers(1, 3), min_size=d, max_size=d).map(tuple)
    )


def real_arrays(shape=None, min_value=None, max_value=None):
    return arrays(
        dtype=np.float64,
        shape=shape if shape is not None else valid_shapes(),
        elements=st.floats(
            min_value=min_value,
            max_value=max_value,
            allow_nan=False,
            allow_infinity=False,
            width=64,
        ),
    ).map(jnp.array)


def lower_triangular_matrices(n=2):
    return arrays(
        dtype=np.float64,
        shape=(n, n),
        elements=st.floats(min_value=-2.0, max_value=2.0, width=64),
    ).map(lambda x: jnp.tril(jnp.array(x)))


class FlexibleMockModel(nnx.Module):
    def __init__(
        self,
        pos_val,
        real_val,
        non_neg_val,
        sigmoid_val,
        lower_val,
        vec_val,
        pos_prior=None,
        real_prior=None,
        non_neg_prior=None,
        sigmoid_prior=None,
        lower_prior=None,
        vec_prior=None,
    ):
        self.pos = PositiveReal(pos_val, prior=pos_prior)
        self.real = Real(real_val, prior=real_prior)
        self.non_neg = NonNegativeReal(non_neg_val, prior=non_neg_prior)
        self.sigmoid = SigmoidBounded(sigmoid_val, prior=sigmoid_prior)
        self.lower = LowerTriangular(lower_val, prior=lower_prior)
        self.vec = Real(vec_val, prior=vec_prior)


@given(
    pos_val=real_arrays(shape=(1,), min_value=1e-3, max_value=10.0),
    real_val=real_arrays(shape=(1,), min_value=-10.0, max_value=10.0),
    non_neg_val=real_arrays(shape=(1,), min_value=0.0, max_value=10.0),
    sigmoid_val=real_arrays(shape=(1,), min_value=1e-3, max_value=0.999),
    lower_val=lower_triangular_matrices(n=2),
    vec_val=real_arrays(shape=(2,), min_value=-10.0, max_value=10.0),
)
def test_no_priors_no_sampling(
    pos_val, real_val, non_neg_val, sigmoid_val, lower_val, vec_val
):
    model = FlexibleMockModel(
        pos_val, real_val, non_neg_val, sigmoid_val, lower_val, vec_val
    )

    def model_fn():
        return register_parameters(model)

    with seed(rng_seed=0):
        tr = trace(model_fn).get_trace()

    # Should be empty because no priors were attached or passed
    assert len(tr) == 0


@given(
    pos_val=real_arrays(shape=(1,), min_value=1e-3, max_value=10.0),
    real_val=real_arrays(shape=(1,), min_value=-10.0, max_value=10.0),
    non_neg_val=real_arrays(shape=(1,), min_value=0.0, max_value=10.0),
    sigmoid_val=real_arrays(shape=(1,), min_value=1e-3, max_value=0.999),
    lower_val=lower_triangular_matrices(n=2),
    vec_val=real_arrays(shape=(2,), min_value=-10.0, max_value=10.0),
)
def test_explicit_priors_sampling(
    pos_val, real_val, non_neg_val, sigmoid_val, lower_val, vec_val
):
    model = FlexibleMockModel(
        pos_val, real_val, non_neg_val, sigmoid_val, lower_val, vec_val
    )

    # Define priors compatible with shapes
    priors = {
        "pos": dist.LogNormal(0.0, 1.0).expand(pos_val.shape).to_event(pos_val.ndim),
        "real": dist.Normal(0.0, 1.0).expand(real_val.shape).to_event(real_val.ndim),
        "non_neg": dist.LogNormal(0.0, 1.0)
        .expand(non_neg_val.shape)
        .to_event(non_neg_val.ndim),
        "sigmoid": dist.Uniform(0.0, 1.0)
        .expand(sigmoid_val.shape)
        .to_event(sigmoid_val.ndim),
        # For LowerTriangular, user must provide a prior over the full matrix shape
        # OR a transformed prior. Here we simulate providing a prior over the full shape
        # just to ensure the site is registered.
        "lower": dist.Normal(0.0, 1.0).expand(lower_val.shape).to_event(lower_val.ndim),
        "vec": dist.Normal(0.0, 1.0).expand(vec_val.shape).to_event(vec_val.ndim),
    }

    def model_fn():
        return register_parameters(model, priors=priors)

    with seed(rng_seed=0):
        tr = trace(model_fn).get_trace()

    assert "pos" in tr
    assert "real" in tr
    assert "non_neg" in tr
    assert "sigmoid" in tr
    assert "lower" in tr
    assert "vec" in tr


@given(
    pos_val=real_arrays(shape=(1,), min_value=1e-3, max_value=10.0),
    real_val=real_arrays(shape=(1,), min_value=-10.0, max_value=10.0),
    non_neg_val=real_arrays(shape=(1,), min_value=0.0, max_value=10.0),
    sigmoid_val=real_arrays(shape=(1,), min_value=1e-3, max_value=0.999),
    lower_val=lower_triangular_matrices(n=2),
    vec_val=real_arrays(shape=(2,), min_value=-10.0, max_value=10.0),
)
def test_attached_priors_sampling(
    pos_val, real_val, non_neg_val, sigmoid_val, lower_val, vec_val
):
    # Create priors
    pos_prior = dist.LogNormal(0.0, 1.0).expand(pos_val.shape).to_event(pos_val.ndim)
    real_prior = dist.Normal(0.0, 1.0).expand(real_val.shape).to_event(real_val.ndim)
    # Attach only to a subset to verify mixed behavior
    model = FlexibleMockModel(
        pos_val,
        real_val,
        non_neg_val,
        sigmoid_val,
        lower_val,
        vec_val,
        pos_prior=pos_prior,
        real_prior=real_prior,
    )

    def model_fn():
        return register_parameters(model)

    with seed(rng_seed=0):
        tr = trace(model_fn).get_trace()

    assert "pos" in tr
    assert "real" in tr
    assert "non_neg" not in tr
    assert "vec" not in tr


@given(
    pos_val=real_arrays(shape=(1,), min_value=1e-3, max_value=10.0),
)
def test_prior_precedence(pos_val):
    # Attached prior
    attached_prior = dist.Gamma(2.0, 1.0).expand(pos_val.shape).to_event(pos_val.ndim)

    # Explicit prior (different)
    explicit_prior = dist.Exponential(1.0).expand(pos_val.shape).to_event(pos_val.ndim)

    # Model with attached prior
    # We need dummy values for others
    dummy_real = jnp.array([0.0])
    dummy_lower = jnp.eye(2)
    dummy_vec = jnp.zeros(2)

    model = FlexibleMockModel(
        pos_val,
        dummy_real,
        dummy_real,
        dummy_real,
        dummy_lower,
        dummy_vec,
        pos_prior=attached_prior,
    )

    priors = {"pos": explicit_prior}

    def model_fn():
        return register_parameters(model, priors=priors)

    with seed(rng_seed=0):
        tr = trace(model_fn).get_trace()

    # Check that the sampled site corresponds to the explicit prior
    # We can check the distribution object
    # Structure might be Independent(Expanded(Exponential)) or Independent(Exponential)
    d = tr["pos"]["fn"]
    while hasattr(d, "base_dist"):
        d = d.base_dist
    assert isinstance(d, dist.Exponential)


def test_register_parameters_nested_prefix():
    class NestedModel(nnx.Module):
        def __init__(self):
            self.inner = FlexibleMockModel(
                jnp.array([1.0]),
                jnp.array([0.0]),
                jnp.array([1.0]),
                jnp.array([0.5]),
                jnp.eye(2),
                jnp.zeros(2),
            )

    model = NestedModel()
    # Explicit prior for nested
    priors = {"outer.inner.pos": dist.LogNormal(0.0, 1.0).expand((1,)).to_event(1)}

    def model_fn():
        return register_parameters(model, prefix="outer", priors=priors)

    with seed(rng_seed=0):
        tr = trace(model_fn).get_trace()

    assert "outer.inner.pos" in tr
    assert "outer.inner.real" not in tr


def test_tree_path_to_name_with_getattr_key():
    path = (jtu.GetAttrKey("kernel"), jtu.GetAttrKey("lengthscale"))
    assert tree_path_to_name(path) == "kernel.lengthscale"


def test_tree_path_to_name_with_dict_key():
    path = (jtu.DictKey(key="params"), jtu.DictKey(key="variance"))
    assert tree_path_to_name(path) == "params.variance"


def test_tree_path_to_name_with_sequence_key():
    path = (jtu.GetAttrKey("layers"), jtu.SequenceKey(idx=0), jtu.GetAttrKey("weight"))
    assert tree_path_to_name(path) == "layers.0.weight"


def test_tree_path_to_name_with_prefix():
    path = (jtu.GetAttrKey("kernel"), jtu.GetAttrKey("variance"))
    assert tree_path_to_name(path, prefix="model") == "model.kernel.variance"


def test_tree_path_to_name_empty_path():
    path = ()
    assert tree_path_to_name(path) == ""


def test_tree_path_to_name_empty_path_with_prefix():
    path = ()
    assert tree_path_to_name(path, prefix="model") == "model."


def test_tree_path_to_name_mixed_keys():
    path = (
        jtu.GetAttrKey("nested"),
        jtu.DictKey(key="sub"),
        jtu.SequenceKey(idx=2),
    )
    assert tree_path_to_name(path) == "nested.sub.2"


def test_resolve_prior_explicit_takes_precedence():
    explicit_prior = dist.Normal(0.0, 1.0)
    attached_prior = dist.Gamma(1.0, 1.0)
    param = PositiveReal(jnp.array([1.0]), prior=attached_prior)
    priors = {"my_param": explicit_prior}

    result = resolve_prior("my_param", param, priors)
    assert result is explicit_prior


def test_resolve_prior_falls_back_to_attached():
    attached_prior = dist.LogNormal(0.0, 1.0)
    param = PositiveReal(jnp.array([1.0]), prior=attached_prior)
    priors = {}

    result = resolve_prior("my_param", param, priors)
    assert result is attached_prior


def test_resolve_prior_returns_none_when_no_prior():
    param = Real(jnp.array([0.0]))
    priors = {}

    result = resolve_prior("my_param", param, priors)
    assert result is None


def test_resolve_prior_explicit_for_different_name_no_attached():
    explicit_prior = dist.Normal(0.0, 1.0)
    param = Real(jnp.array([0.0]))
    priors = {"other_param": explicit_prior}

    result = resolve_prior("my_param", param, priors)
    assert result is None


def test_register_parameters_conjugate_posterior():
    """Integration test: register_parameters on a real ConjugatePosterior.

    Verifies that:
    - Nested modules (kernel, likelihood) are traversed correctly
    - Shared nnx.Variable references (lengthscale shared between RBF and Periodic)
      result in a single sample site
    - nnx.List inside CombinationKernel is traversed properly
    - All parameters with priors are sampled
    - conjugate_mll can be evaluated with the sampled parameters
    """
    import gpjax as gpx

    lengthscale_prior = dist.LogNormal(0.0, 1.0)
    variance_prior = dist.LogNormal(0.0, 1.0)
    period_prior = dist.LogNormal(0.0, 0.5)
    noise_prior = dist.LogNormal(0.0, 1.0)

    lengthscale = PositiveReal(1.0, prior=lengthscale_prior)
    variance = PositiveReal(1.0, prior=variance_prior)
    period = PositiveReal(1.0, prior=period_prior)
    noise = NonNegativeReal(1.0, prior=noise_prior)

    rbf = gpx.kernels.RBF(lengthscale=lengthscale, variance=variance)
    periodic = gpx.kernels.Periodic(lengthscale=lengthscale, period=period)
    kernel = rbf * periodic

    meanf = gpx.mean_functions.Constant()
    prior = gpx.gps.Prior(mean_function=meanf, kernel=kernel)
    likelihood = gpx.likelihoods.Gaussian(num_datapoints=10, obs_stddev=noise)
    posterior = prior * likelihood

    def model_fn():
        return register_parameters(posterior)

    with seed(rng_seed=42):
        tr = trace(model_fn).get_trace()

    # Shared lengthscale should appear once (shared Variable → single site)
    lengthscale_sites = [k for k in tr if "lengthscale" in k]
    assert len(lengthscale_sites) == 1, (
        f"Expected 1 lengthscale site, got {len(lengthscale_sites)}: {lengthscale_sites}"
    )

    # variance, period, and obs_stddev should each appear once
    variance_sites = [k for k in tr if "variance" in k]
    assert len(variance_sites) == 1

    period_sites = [k for k in tr if "period" in k]
    assert len(period_sites) == 1

    obs_stddev_sites = [k for k in tr if "obs_stddev" in k]
    assert len(obs_stddev_sites) == 1

    # Total: 4 sampled sites (lengthscale, variance, period, obs_stddev)
    assert len(tr) == 4, f"Expected 4 sample sites, got {len(tr)}: {list(tr.keys())}"


def test_register_parameters_conjugate_posterior_mll():
    """Integration test: sampled parameters flow through conjugate_mll."""
    import gpjax as gpx
    import jax.random as jr

    key = jr.key(0)
    X = jr.uniform(key, shape=(10, 1))
    y = jnp.sin(X)
    D = gpx.Dataset(X=X, y=y)

    lengthscale = PositiveReal(1.0, prior=dist.LogNormal(0.0, 1.0))
    variance = PositiveReal(1.0, prior=dist.LogNormal(0.0, 1.0))
    noise = NonNegativeReal(0.5, prior=dist.LogNormal(0.0, 1.0))

    kernel = gpx.kernels.RBF(lengthscale=lengthscale, variance=variance)
    meanf = gpx.mean_functions.Constant()
    prior = gpx.gps.Prior(mean_function=meanf, kernel=kernel)
    likelihood = gpx.likelihoods.Gaussian(num_datapoints=10, obs_stddev=noise)
    posterior = prior * likelihood

    mll_value = None

    def model_fn():
        nonlocal mll_value
        p = register_parameters(posterior)
        mll_value = gpx.objectives.conjugate_mll(p, D)
        return mll_value

    with seed(rng_seed=42):
        tr = trace(model_fn).get_trace()

    assert len(tr) == 3  # lengthscale, variance, obs_stddev
    assert mll_value is not None
    assert jnp.isfinite(mll_value)



# Copyright 2023 The thomaspinder Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import typing as tp

from flax import nnx
import jax
from jax.flatten_util import ravel_pytree
import jax.numpy as jnp
import jax.random as jr
from numpyro.distributions.transforms import Transform
import optax as ox
from scipy.optimize import minimize

from gpjax.dataset import Dataset
from gpjax.objectives import Objective
from gpjax.parameters import (
    DEFAULT_BIJECTION,
    Parameter,
    transform,
)
from gpjax.scan import vscan
from gpjax.typing import (
    Array,
    KeyArray,
    ScalarFloat,
)

Model = tp.TypeVar("Model", bound=nnx.Module)


def fit(
    *,
    model: Model,
    objective: Objective,
    train_data: Dataset,
    optim: ox.GradientTransformation,
    params_bijection: dict[Parameter, Transform] | None = DEFAULT_BIJECTION,
    trainable: nnx.filterlib.Filter = Parameter,
    key: KeyArray = jr.key(42),
    num_iters: int = 100,
    batch_size: int = -1,
    log_rate: int = 10,
    verbose: bool = True,
    unroll: int = 1,
    safe: bool = True,
) -> tuple[Model, jax.Array]:
    r"""Train a Module model with respect to a supplied objective function.
    Optimisers used here should originate from Optax.

    Example:
    ```pycon
        >>> import jax.numpy as jnp
        >>> import jax.random as jr
        >>> import optax as ox
        >>> import gpjax as gpx
        >>> from gpjax.parameters import PositiveReal
        >>>
        >>> # (1) Create a dataset:
        >>> X = jnp.linspace(0.0, 10.0, 100)[:, None]
        >>> y = 2.0 * X + 1.0 + 10 * jr.normal(jr.key(0), X.shape)
        >>> D = gpx.Dataset(X, y)
        >>> # (2) Define your model:
        >>> class LinearModel(nnx.Module):
        >>>     def __init__(self, weight: float, bias: float):
        >>>         self.weight = PositiveReal(weight)
        >>>         self.bias = bias
        >>>
        >>>     def __call__(self, x):
        >>>         return self.weight[...] * x + self.bias
        >>>
        >>> model = LinearModel(weight=1.0, bias=1.0)
        >>>
        >>> # (3) Define your loss function:
        >>> def mse(model, data):
        >>>     pred = model(data.X)
        >>>     return jnp.mean((pred - data.y) ** 2)
        >>>
        >>> # (4) Train!
        >>> trained_model, history = gpx.fit(
        >>>     model=model, objective=mse, train_data=D, optim=ox.sgd(0.001), num_iters=1000
        >>> )
    ```

    Args:
        model (Model): The model Module to be optimised.
        objective (Objective): The objective function that we are optimising with
            respect to.
        train_data (Dataset): The training data to be used for the optimisation.
        optim (GradientTransformation): The Optax optimiser that is to be used for
            learning a parameter set.
        trainable (nnx.filterlib.Filter): Filter to determine which parameters are trainable.
            Defaults to nnx.Param (all Parameter instances).
        num_iters (int): The number of optimisation steps to run. Defaults
            to 100.
        batch_size (int): The size of the mini-batch to use. Defaults to -1
            (i.e. full batch).
        key (KeyArray): The random key to use for the optimisation batch
            selection. Defaults to jr.key(42).
        log_rate (int): How frequently the objective function's value should
            be printed. Defaults to 10.
        verbose (bool): Whether to print the training loading bar. Defaults
            to True.
        unroll (int): The number of unrolled steps to use for the optimisation.
            Defaults to 1.

    Returns:
        A tuple comprising the optimised model and training history.
    """
    if safe:
        # Check inputs.
        _check_model(model)
        _check_train_data(train_data)
        _check_optim(optim)
        _check_num_iters(num_iters)
        _check_batch_size(batch_size)
        _check_log_rate(log_rate)
        _check_verbose(verbose)

    # Model state filtering
    graphdef, params, *static_state = nnx.split(model, trainable, ...)

    # Parameters bijection to unconstrained space
    if params_bijection is not None:
        params = transform(params, params_bijection, inverse=True)

    # Loss definition
    def loss(params: nnx.State, batch: Dataset) -> ScalarFloat:
        params = transform(params, params_bijection)
        model = nnx.merge(graphdef, params, *static_state)
        return objective(model, batch)

    # Initialise optimiser state.
    opt_state = optim.init(params)

    # Mini-batch random keys to scan over.
    iter_keys = jr.split(key, num_iters)

    # Optimisation step.
    def step(carry, key):
        params, opt_state = carry

        if batch_size != -1:
            batch = get_batch(train_data, batch_size, key)
        else:
            batch = train_data

        loss_val, loss_gradient = jax.value_and_grad(loss)(params, batch)
        updates, opt_state = optim.update(loss_gradient, opt_state, params)
        params = ox.apply_updates(params, updates)

        carry = params, opt_state
        return carry, loss_val

    # Optimisation scan.
    scan = vscan if verbose else jax.lax.scan

    # Optimisation loop.
    (params, _), history = scan(step, (params, opt_state), (iter_keys), unroll=unroll)

    # Parameters bijection to constrained space
    if params_bijection is not None:
        params = transform(params, params_bijection)

    # Reconstruct model
    model = nnx.merge(graphdef, params, *static_state)

    return model, history


def fit_scipy(
    *,
    model: Model,
    objective: Objective,
    train_data: Dataset,
    trainable: nnx.filterlib.Filter = Parameter,
    max_iters: int = 500,
    verbose: bool = True,
    safe: bool = True,
) -> tuple[Model, Array]:
    r"""Train a Module model with respect to a supplied Objective function
    using SciPy's L-BFGS-B optimiser.

    Parameters are transformed to unconstrained space, flattened into a
    single vector, and passed to ``scipy.optimize.minimize``. Gradients
    are computed via JAX's ``value_and_grad``.

    Parameters
    ----------
    model : Module
        The model to be optimised.
    objective : Objective
        The objective function to minimise with respect to the model
        parameters.
    train_data : Dataset
        The training data used to evaluate the objective.
    trainable : nnx.filterlib.Filter
        Filter selecting which parameters to optimise. Defaults to all
        ``Parameter`` instances.
    max_iters : int
        Maximum number of L-BFGS-B iterations. Defaults to 500.
    verbose : bool
        Whether to print optimisation progress. Defaults to True.
    safe : bool
        Whether to validate inputs before optimisation. Defaults to True.

    Returns
    -------
    tuple[Module, Array]
        A tuple of the optimised model and an array of objective values
        recorded at each iteration.

    Example:
        >>> import gpjax as gpx
        >>> import jax.numpy as jnp

        >>> xtrain = jnp.linspace(0, 1).reshape(-1, 1)
        >>> ytrain = jnp.sin(xtrain)
        >>> D = gpx.Dataset(X=xtrain, y=ytrain)

        >>> meanf = gpx.mean_functions.Constant()
        >>> kernel = gpx.kernels.RBF()
        >>> likelihood = gpx.likelihoods.Gaussian(num_datapoints=D.n)
        >>> prior = gpx.gps.Prior(mean_function=meanf, kernel=kernel)
        >>> posterior = prior * likelihood

        >>> nmll = lambda p, d: -gpx.objectives.conjugate_mll(p, d)
        >>> trained_model, history = gpx.fit_scipy(
        ...     model=posterior, objective=nmll, train_data=D
        ... )
    """
    if safe:
        # Check inputs.
        _check_model(model)
        _check_train_data(train_data)
        _check_num_iters(max_iters)
        _check_verbose(verbose)

    # Model state filtering
    graphdef, params, *static_state = nnx.split(model, trainable, ...)

    # Parameters bijection to unconstrained space
    params = transform(params, DEFAULT_BIJECTION, inverse=True)

    # Loss definition
    def loss(params) -> ScalarFloat:
        params = transform(params, DEFAULT_BIJECTION)
        model = nnx.merge(graphdef, params, *static_state)
        return objective(model, train_data)

    # convert to numpy for interface with scipy
    x0, scipy_to_jnp = ravel_pytree(params)

    @jax.jit
    def scipy_wrapper(x0):
        value, grads = jax.value_and_grad(loss)(scipy_to_jnp(jnp.array(x0)))
        scipy_grads = ravel_pytree(grads)[0]
        return value, scipy_grads

    history = [scipy_wrapper(x0)[0]]
    result = minimize(
        fun=scipy_wrapper,
        x0=x0,
        jac=True,
        callback=lambda X: history.append(scipy_wrapper(X)[0]),
        options={"maxiter": max_iters, "disp": verbose},
    )
    history = jnp.array(history)

    # convert back to nnx.State with JAX arrays
    params = scipy_to_jnp(result.x)

    # Parameters bijection to constrained space
    params = transform(params, DEFAULT_BIJECTION)

    # Reconstruct model
    model = nnx.merge(graphdef, params, *static_state)

    return model, history


def fit_lbfgs(
    *,
    model: Model,
    objective: Objective,
    train_data: Dataset,
    params_bijection: dict[Parameter, Transform] | None = DEFAULT_BIJECTION,
    trainable: nnx.filterlib.Filter = Parameter,
    max_iters: int = 100,
    safe: bool = True,
    max_linesearch_steps: int = 32,
    gtol: float = 1e-5,
) -> tuple[Model, jax.Array]:
    r"""Train a Module model with respect to a supplied Objective function.

    Uses Optax's L-BFGS implementation with a ``jax.lax.while_loop``.

    Parameters
    ----------
    model : Module
        The model to be optimised.
    objective : Objective
        The objective function to minimise.
    train_data : Dataset
        The training data used to evaluate the objective.
    params_bijection : dict[Parameter, Transform] | None
        Bijection used to transform parameters to unconstrained space.
        Defaults to ``DEFAULT_BIJECTION``.
    trainable : nnx.filterlib.Filter
        Filter selecting which parameters to optimise. Defaults to all
        ``Parameter`` instances.
    max_iters : int
        Maximum number of L-BFGS iterations. Defaults to 100.
    safe : bool
        Whether to validate inputs before optimisation. Defaults to True.
    max_linesearch_steps : int
        Maximum number of line-search steps per iteration. Defaults to 32.
    gtol : float
        Terminate if the L2 norm of the gradient falls below this
        threshold. Defaults to 1e-5.

    Returns
    -------
    tuple[Module, Array]
        A tuple of the optimised model and the final loss value.

    Example:
        >>> import gpjax as gpx
        >>> import jax.numpy as jnp

        >>> xtrain = jnp.linspace(0, 1).reshape(-1, 1)
        >>> ytrain = jnp.sin(xtrain)
        >>> D = gpx.Dataset(X=xtrain, y=ytrain)

        >>> meanf = gpx.mean_functions.Constant()
        >>> kernel = gpx.kernels.RBF()
        >>> likelihood = gpx.likelihoods.Gaussian(num_datapoints=D.n)
        >>> prior = gpx.gps.Prior(mean_function=meanf, kernel=kernel)
        >>> posterior = prior * likelihood

        >>> nmll = lambda p, d: -gpx.objectives.conjugate_mll(p, d)
        >>> trained_model, final_loss = gpx.fit_lbfgs(
        ...     model=posterior, objective=nmll, train_data=D
        ... )
    """
    if safe:
        # Check inputs
        _check_model(model)
        _check_train_data(train_data)
        _check_num_iters(max_iters)

    # Model state filtering
    graphdef, params, *static_state = nnx.split(model, trainable, ...)

    # Parameters bijection to unconstrained space
    if params_bijection is not None:
        params = transform(params, params_bijection, inverse=True)

    # Loss definition
    def loss(params: nnx.State) -> ScalarFloat:
        params = transform(params, params_bijection)
        model = nnx.merge(graphdef, params, *static_state)
        return objective(model, train_data)

    # Initialise optimiser
    optim = ox.lbfgs(
        linesearch=ox.scale_by_zoom_linesearch(
            max_linesearch_steps=max_linesearch_steps,
            initial_guess_strategy="one",
        )
    )
    opt_state = optim.init(params)
    loss_value_and_grad = ox.value_and_grad_from_state(loss)

    # Optimisation step.
    def step(carry):
        params, opt_state = carry

        # Using optax's value_and_grad_from_state is more efficient given LBFGS uses a linesearch
        # See https://optax.readthedocs.io/en/latest/api/utilities.html#optax.value_and_grad_from_state
        loss_val, loss_gradient = loss_value_and_grad(params, state=opt_state)
        updates, opt_state = optim.update(
            loss_gradient,
            opt_state,
            params,
            value=loss_val,
            grad=loss_gradient,
            value_fn=loss,
        )
        params = ox.apply_updates(params, updates)

        return params, opt_state

    def continue_fn(carry):
        _, opt_state = carry
        n = ox.tree_utils.tree_get(opt_state, "count")
        g = ox.tree_utils.tree_get(opt_state, "grad")
        g_l2_norm = ox.tree_utils.tree_l2_norm(g)
        return (n == 0) | ((n < max_iters) & (g_l2_norm >= gtol))

    # Optimisation loop
    params, opt_state = jax.lax.while_loop(
        continue_fn,
        step,
        (params, opt_state),
    )
    final_loss = ox.tree_utils.tree_get(opt_state, "value")

    # Parameters bijection to constrained space
    if params_bijection is not None:
        params = transform(params, params_bijection)

    # Reconstruct model
    model = nnx.merge(graphdef, params, *static_state)

    return model, final_loss


def get_batch(train_data: Dataset, batch_size: int, key: KeyArray) -> Dataset:
    """Batch the data into mini-batches. Sampling is done with replacement.

    Args:
        train_data (Dataset): The training dataset.
        batch_size (int): The batch size.
        key (KeyArray): The random key to use for the batch selection.

    Example:
        >>> import gpjax as gpx
        >>> import jax.numpy as jnp
        >>> import jax.random as jr

        >>> X = jnp.linspace(0, 1, 100).reshape(-1, 1)
        >>> y = jnp.sin(X)
        >>> D = gpx.Dataset(X=X, y=y)

        >>> from gpjax.fit import get_batch
        >>> batch = get_batch(D, batch_size=16, key=jr.key(0))

    Returns
    -------
    Dataset
        The batched dataset.
    """
    x, y, n = train_data.X, train_data.y, train_data.n

    # Subsample mini-batch indices with replacement.
    indices = jr.choice(key, n, (batch_size,), replace=True)

    return Dataset(X=x[indices], y=y[indices])


def _check_model(model: tp.Any) -> None:
    """Check that the model is a subclass of nnx.Module."""
    if not isinstance(model, nnx.Module):
        raise TypeError(
            "Expected model to be a subclass of nnx.Module. "
            f"Got {model} of type {type(model)}."
        )


def _check_train_data(train_data: tp.Any) -> None:
    """Check that the train_data is of type gpjax.Dataset."""
    if not isinstance(train_data, Dataset):
        raise TypeError(
            "Expected train_data to be of type gpjax.Dataset. "
            f"Got {train_data} of type {type(train_data)}."
        )


def _check_optim(optim: tp.Any) -> None:
    """Check that the optimiser is of type GradientTransformation."""
    if not isinstance(optim, ox.GradientTransformation):
        raise TypeError(
            "Expected optim to be of type optax.GradientTransformation. "
            f"Got {optim} of type {type(optim)}."
        )


def _check_num_iters(num_iters: tp.Any) -> None:
    """Check that the number of iterations is of type int and positive."""
    if not isinstance(num_iters, int):
        raise TypeError(
            "Expected num_iters to be of type int. "
            f"Got {num_iters} of type {type(num_iters)}."
        )

    if num_iters <= 0:
        raise ValueError(f"Expected num_iters to be positive. Got {num_iters}.")


def _check_log_rate(log_rate: tp.Any) -> None:
    """Check that the log rate is of type int and positive."""
    if not isinstance(log_rate, int):
        raise TypeError(
            "Expected log_rate to be of type int. "
            f"Got {log_rate} of type {type(log_rate)}."
        )

    if not log_rate > 0:
        raise ValueError(f"Expected log_rate to be positive. Got {log_rate}.")


def _check_verbose(verbose: tp.Any) -> None:
    """Check that the verbose is of type bool."""
    if not isinstance(verbose, bool):
        raise TypeError(
            "Expected verbose to be of type bool. "
            f"Got {verbose} of type {type(verbose)}."
        )


def _check_batch_size(batch_size: tp.Any) -> None:
    """Check that the batch size is of type int and positive if not minus 1."""
    if not isinstance(batch_size, int):
        raise TypeError(
            "Expected batch_size to be of type int. "
            f"Got {batch_size} of type {type(batch_size)}."
        )

    if not batch_size == -1 and not batch_size > 0:
        raise ValueError(f"Expected batch_size to be positive or -1. Got {batch_size}.")


__all__ = [
    "fit",
    "fit_lbfgs",
    "fit_scipy",
    "get_batch",
]



"""Tests for numerical stability of kernel computations and GP predictions.

Parametrized across kernel types, these tests exercise edge cases that may
trigger NaN/Inf: very small lengthscales, very large variances, extreme input
ranges, and near-singular kernel matrices.
"""

import gpjax as gpx
from gpjax.kernels.stationary import (
    RBF,
    Matern12,
    Matern32,
    Matern52,
    RationalQuadratic,
)
from gpjax.linalg.utils import add_jitter
from gpjax.mean_functions import Zero
import jax.numpy as jnp
import jax.random as jr
import pytest

KERNEL_CLASSES = [RBF, Matern12, Matern32, Matern52, RationalQuadratic]


def _make_kernel(kernel_cls, lengthscale=1.0, variance=1.0):
    """Create a kernel instance with the given parameters."""
    return kernel_cls(lengthscale=lengthscale, variance=variance)


# ---------------------------------------------------------------------------
# Kernel gram matrix stability
# ---------------------------------------------------------------------------


class TestKernelGramStability:
    """Test that kernel gram matrices remain finite under extreme parameters."""

    @pytest.mark.parametrize("kernel_cls", KERNEL_CLASSES)
    def test_small_lengthscale(self, kernel_cls):
        """Very small lengthscale should produce finite gram matrix."""
        kernel = _make_kernel(kernel_cls, lengthscale=1e-6)
        x = jnp.linspace(0.0, 1.0, 10).reshape(-1, 1)
        gram = kernel.gram(x).to_dense()
        assert jnp.all(jnp.isfinite(gram)), (
            f"Non-finite values in gram with small lengthscale for {kernel_cls.__name__}"
        )

    @pytest.mark.parametrize("kernel_cls", KERNEL_CLASSES)
    def test_large_variance(self, kernel_cls):
        """Large variance should produce finite gram matrix."""
        kernel = _make_kernel(kernel_cls, variance=1e6)
        x = jnp.linspace(0.0, 1.0, 10).reshape(-1, 1)
        gram = kernel.gram(x).to_dense()
        assert jnp.all(jnp.isfinite(gram)), (
            f"Non-finite values in gram with large variance for {kernel_cls.__name__}"
        )

    @pytest.mark.parametrize("kernel_cls", KERNEL_CLASSES)
    def test_large_input_range(self, kernel_cls):
        """Extreme input ranges should produce finite gram matrix."""
        kernel = _make_kernel(kernel_cls)
        x = jnp.linspace(-1e4, 1e4, 10).reshape(-1, 1)
        gram = kernel.gram(x).to_dense()
        assert jnp.all(jnp.isfinite(gram)), (
            f"Non-finite values in gram with large inputs for {kernel_cls.__name__}"
        )

    @pytest.mark.parametrize("kernel_cls", KERNEL_CLASSES)
    def test_identical_points(self, kernel_cls):
        """Identical input points should produce finite gram matrix."""
        kernel = _make_kernel(kernel_cls)
        x = jnp.ones((5, 1))
        gram = kernel.gram(x).to_dense()
        assert jnp.all(jnp.isfinite(gram))

    @pytest.mark.parametrize("kernel_cls", KERNEL_CLASSES)
    def test_very_close_points(self, kernel_cls):
        """Very close but distinct points should produce finite gram matrix."""
        kernel = _make_kernel(kernel_cls)
        x = jnp.array([[0.0], [1e-10], [2e-10], [3e-10], [4e-10]])
        gram = kernel.gram(x).to_dense()
        assert jnp.all(jnp.isfinite(gram))


# ---------------------------------------------------------------------------
# Cholesky stability with jitter
# ---------------------------------------------------------------------------


class TestCholeskyStability:
    """Test that Cholesky decomposition succeeds with default jitter."""

    @pytest.mark.parametrize("kernel_cls", KERNEL_CLASSES)
    def test_cholesky_with_jitter(self, kernel_cls):
        """Cholesky should succeed on jittered kernel gram matrix."""
        kernel = _make_kernel(kernel_cls)
        x = jnp.linspace(0.0, 1.0, 20).reshape(-1, 1)
        gram = kernel.gram(x).to_dense()
        jittered = add_jitter(gram, jitter=1e-6)
        L = jnp.linalg.cholesky(jittered)
        assert jnp.all(jnp.isfinite(L)), f"Cholesky failed for {kernel_cls.__name__}"

    @pytest.mark.parametrize("kernel_cls", KERNEL_CLASSES)
    def test_cholesky_identical_points(self, kernel_cls):
        """Cholesky should succeed on near-singular gram (identical points + jitter)."""
        kernel = _make_kernel(kernel_cls)
        x = jnp.ones((10, 1))
        gram = kernel.gram(x).to_dense()
        jittered = add_jitter(gram, jitter=1e-6)
        L = jnp.linalg.cholesky(jittered)
        assert jnp.all(jnp.isfinite(L))


# ---------------------------------------------------------------------------
# GP posterior prediction stability
# ---------------------------------------------------------------------------


class TestPosteriorPredictionStability:
    """Test that full GP pipeline (prior -> posterior -> predict) stays finite."""

    @pytest.mark.parametrize("kernel_cls", KERNEL_CLASSES)
    def test_standard_predict_finite(self, kernel_cls):
        """Standard prediction should produce finite mean and covariance."""
        kernel = _make_kernel(kernel_cls)
        prior = gpx.gps.Prior(mean_function=Zero(), kernel=kernel)
        likelihood = gpx.likelihoods.Gaussian(num_datapoints=20)
        posterior = prior * likelihood

        x_train = jnp.linspace(0.0, 1.0, 20).reshape(-1, 1)
        y_train = jnp.sin(2 * jnp.pi * x_train)
        D = gpx.Dataset(X=x_train, y=y_train)

        x_test = jnp.linspace(0.0, 1.0, 10).reshape(-1, 1)
        pred_dist = posterior.predict(x_test, D)

        mean = pred_dist.mean
        cov = pred_dist.covariance()
        assert jnp.all(jnp.isfinite(mean)), f"Non-finite mean for {kernel_cls.__name__}"
        assert jnp.all(jnp.isfinite(cov)), (
            f"Non-finite covariance for {kernel_cls.__name__}"
        )

    @pytest.mark.parametrize("kernel_cls", KERNEL_CLASSES)
    def test_predict_with_noisy_data(self, kernel_cls):
        """Prediction with noisy data should stay finite."""
        kernel = _make_kernel(kernel_cls)
        prior = gpx.gps.Prior(mean_function=Zero(), kernel=kernel)
        likelihood = gpx.likelihoods.Gaussian(num_datapoints=30)
        posterior = prior * likelihood

        key = jr.key(42)
        x_train = jnp.linspace(-2.0, 2.0, 30).reshape(-1, 1)
        y_train = jnp.sin(x_train) + 0.5 * jr.normal(key, x_train.shape)
        D = gpx.Dataset(X=x_train, y=y_train)

        x_test = jnp.linspace(-3.0, 3.0, 15).reshape(-1, 1)
        pred_dist = posterior.predict(x_test, D)

        assert jnp.all(jnp.isfinite(pred_dist.mean))
        assert jnp.all(jnp.isfinite(pred_dist.covariance()))

    @pytest.mark.parametrize("kernel_cls", KERNEL_CLASSES)
    def test_predict_extrapolation(self, kernel_cls):
        """Predictions far from training data should stay finite."""
        kernel = _make_kernel(kernel_cls)
        prior = gpx.gps.Prior(mean_function=Zero(), kernel=kernel)
        likelihood = gpx.likelihoods.Gaussian(num_datapoints=10)
        posterior = prior * likelihood

        x_train = jnp.linspace(0.0, 1.0, 10).reshape(-1, 1)
        y_train = jnp.cos(x_train)
        D = gpx.Dataset(X=x_train, y=y_train)

        # Test inputs far from training range
        x_test = jnp.linspace(50.0, 100.0, 5).reshape(-1, 1)
        pred_dist = posterior.predict(x_test, D)

        assert jnp.all(jnp.isfinite(pred_dist.mean))
        assert jnp.all(jnp.isfinite(pred_dist.covariance()))

    @pytest.mark.parametrize("kernel_cls", KERNEL_CLASSES)
    def test_diagonal_predict_finite(self, kernel_cls):
        """Diagonal covariance prediction should stay finite."""
        kernel = _make_kernel(kernel_cls)
        prior = gpx.gps.Prior(mean_function=Zero(), kernel=kernel)
        likelihood = gpx.likelihoods.Gaussian(num_datapoints=20)
        posterior = prior * likelihood

        x_train = jnp.linspace(0.0, 1.0, 20).reshape(-1, 1)
        y_train = jnp.sin(x_train)
        D = gpx.Dataset(X=x_train, y=y_train)

        x_test = jnp.linspace(0.0, 1.0, 10).reshape(-1, 1)
        pred_dist = posterior.predict(x_test, D, return_covariance_type="diagonal")

        assert jnp.all(jnp.isfinite(pred_dist.mean))
        assert jnp.all(jnp.isfinite(pred_dist.covariance()))

    @pytest.mark.parametrize("kernel_cls", KERNEL_CLASSES)
    def test_predictive_variance_nonnegative(self, kernel_cls):
        """Diagonal of predictive covariance should be non-negative."""
        kernel = _make_kernel(kernel_cls)
        prior = gpx.gps.Prior(mean_function=Zero(), kernel=kernel)
        likelihood = gpx.likelihoods.Gaussian(num_datapoints=15)
        posterior = prior * likelihood

        x_train = jnp.linspace(0.0, 1.0, 15).reshape(-1, 1)
        y_train = x_train**2
        D = gpx.Dataset(X=x_train, y=y_train)

        x_test = jnp.linspace(-0.5, 1.5, 10).reshape(-1, 1)
        pred_dist = posterior.predict(x_test, D)

        var = jnp.diag(pred_dist.covariance())
        assert jnp.all(var >= -1e-6), (
            f"Negative variance for {kernel_cls.__name__}: {var}"
        )



# Copyright 2023 The thomaspinder Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from beartype.typing import Any
from flax import nnx
import gpjax as gpx
from gpjax.dataset import Dataset
from gpjax.fit import (
    _check_batch_size,
    _check_log_rate,
    _check_model,
    _check_num_iters,
    _check_optim,
    _check_train_data,
    _check_verbose,
    fit,
    fit_lbfgs,
    fit_scipy,
    get_batch,
)
from gpjax.gps import (
    ConjugatePosterior,
    Prior,
)
from gpjax.kernels import RBF
from gpjax.likelihoods import Gaussian
from gpjax.mean_functions import (
    AbstractMeanFunction,
    Constant,
)
from gpjax.objectives import (
    conjugate_mll,
    elbo,
)
from gpjax.parameters import (
    PositiveReal,
)
from gpjax.typing import Array
from gpjax.variational_families import VariationalGaussian
import jax.numpy as jnp
import jax.random as jr
from jaxtyping import (
    Float,
    Num,
)
import optax as ox
import pytest
import scipy


def test_fit_simple() -> None:
    # Create dataset:
    X = jnp.linspace(0.0, 10.0, 100).reshape(-1, 1)
    y = 2.0 * X + 1.0 + 10 * jr.normal(jr.key(0), X.shape).reshape(-1, 1)
    D = Dataset(X, y)

    # Define linear model:

    class LinearModel(nnx.Module):
        def __init__(self, weight: float, bias: float):
            self.weight = PositiveReal(weight)
            self.bias = bias

        def __call__(self, x):
            return self.weight[...] * x + self.bias

    model = LinearModel(weight=1.0, bias=1.0)

    # Define loss function:
    def mse(model, data):
        pred = model(data.X)
        return jnp.mean((pred - data.y) ** 2)

    # Train!
    trained_model, hist = fit(
        model=model,
        objective=mse,
        train_data=D,
        optim=ox.sgd(0.001),
        num_iters=100,
        key=jr.key(123),
    )

    # Ensure we return a history of the correct length
    assert len(hist) == 100

    # Ensure we return a model of the same class
    assert isinstance(trained_model, LinearModel)

    # Test reduction in loss:
    assert mse(trained_model, D) < mse(model, D)

    # Test stop_gradient on bias:
    assert trained_model.bias == 1.0


def test_fit_scipy_simple():
    # Create dataset:
    X = jnp.linspace(0.0, 10.0, 100).reshape(-1, 1)
    y = 2.0 * X + 1.0 + 10 * jr.normal(jr.key(0), X.shape).reshape(-1, 1)
    D = Dataset(X, y)

    # Define linear model:
    class LinearModel(nnx.Module):
        def __init__(self, weight: float, bias: float):
            self.weight = PositiveReal(weight)
            self.bias = bias

        def __call__(self, x):
            return self.weight[...] * x + self.bias

    model = LinearModel(weight=1.0, bias=1.0)

    # Define loss function:
    def mse(model, data):
        pred = model(data.X)
        return jnp.mean((pred - data.y) ** 2)

    # Train with bfgs!
    trained_model, hist = fit_scipy(
        model=model,
        objective=mse,
        train_data=D,
        max_iters=10,
    )

    # Ensure we return a history of the correct length
    assert len(hist) > 2

    # Ensure we return a model of the same class
    assert isinstance(trained_model, LinearModel)

    # Test reduction in loss:
    assert mse(trained_model, D) < mse(model, D)

    # Test stop_gradient on bias:
    assert trained_model.bias == 1.0


def test_fit_lbfgs_simple():
    # Create dataset:
    X = jnp.linspace(0.0, 10.0, 100).reshape(-1, 1)
    y = 2.0 * X + 1.0 + 10 * jr.normal(jr.key(0), X.shape).reshape(-1, 1)
    D = Dataset(X, y)

    # Define linear model:
    class LinearModel(nnx.Module):
        def __init__(self, weight: float, bias: float):
            self.weight = PositiveReal(weight)
            self.bias = bias

        def __call__(self, x):
            return self.weight[...] * x + self.bias

    model = LinearModel(weight=1.0, bias=1.0)

    # Define loss function:
    def mse(model, data):
        pred = model(data.X)
        return jnp.mean((pred - data.y) ** 2)

    # Train with bfgs!
    trained_model, _final_loss = fit_lbfgs(
        model=model,
        objective=mse,
        train_data=D,
        max_iters=10,
    )

    # Ensure we return a model of the same class
    assert isinstance(trained_model, LinearModel)

    # Test reduction in loss:
    assert mse(trained_model, D) < mse(model, D)

    # Test stop_gradient on bias:
    assert trained_model.bias == 1.0


@pytest.mark.parametrize("n_data", [20])
@pytest.mark.parametrize("verbose", [True, False])
def test_fit_gp_regression(n_data: int, verbose: bool) -> None:
    # Create dataset:
    key = jr.key(123)
    x = jnp.sort(
        jr.uniform(key=key, minval=-2.0, maxval=2.0, shape=(n_data, 1)), axis=0
    )
    y = jnp.sin(x) + jr.normal(key=key, shape=x.shape) * 0.1
    D = Dataset(X=x, y=y)

    # Define GP model:
    prior = Prior(kernel=RBF(), mean_function=Constant())
    likelihood = Gaussian(num_datapoints=n_data)
    posterior = prior * likelihood

    # Train!
    trained_model, history = fit(
        model=posterior,
        objective=conjugate_mll,
        train_data=D,
        optim=ox.adam(0.1),
        num_iters=15,
        verbose=verbose,
        key=jr.key(123),
    )

    # Ensure the trained model is a Gaussian process posterior
    assert isinstance(trained_model, ConjugatePosterior)

    # Ensure we return a history of the correct length
    assert len(history) == 15

    # Ensure we reduce the loss
    assert conjugate_mll(trained_model, D) < conjugate_mll(posterior, D)


@pytest.mark.parametrize("n_data", [20])
def test_fit_lbfgs_gp_regression(n_data: int) -> None:
    # Create dataset:
    key = jr.key(123)
    x = jnp.sort(
        jr.uniform(key=key, minval=-2.0, maxval=2.0, shape=(n_data, 1)), axis=0
    )
    y = jnp.sin(x) + jr.normal(key=key, shape=x.shape) * 0.1
    D = Dataset(X=x, y=y)

    # Define GP model:
    prior = Prior(kernel=RBF(), mean_function=Constant())
    likelihood = Gaussian(num_datapoints=n_data)
    posterior = prior * likelihood

    # Train with BFGS!
    trained_model_bfgs, _final_loss = fit_lbfgs(
        model=posterior,
        objective=conjugate_mll,
        train_data=D,
        max_iters=40,
    )

    # Ensure the trained model is a Gaussian process posterior
    assert isinstance(trained_model_bfgs, ConjugatePosterior)

    # Ensure we reduce the loss
    assert conjugate_mll(trained_model_bfgs, D) < conjugate_mll(posterior, D)


def test_fit_scipy_error_raises() -> None:
    # Create dataset:
    D = Dataset(
        X=jnp.array([[0.0]], dtype=jnp.float64), y=jnp.array([[0.0]], dtype=jnp.float64)
    )

    # build crazy mean function so that opt fails
    class CrazyMean(AbstractMeanFunction):
        def __call__(self, x: Num[Array, "N D"]) -> Float[Array, "N O"]:
            return jnp.heaviside(x, 100.0)

    # Define GP model with crazy mean function:
    prior = Prior(kernel=RBF(), mean_function=CrazyMean())
    likelihood = Gaussian(num_datapoints=2)
    posterior = prior * likelihood

    with pytest.raises(scipy.optimize.OptimizeWarning):
        fit_scipy(
            model=posterior,
            objective=conjugate_mll,
            train_data=D,
            max_iters=10,
        )

    # also check fails if no given enough steps
    prior = Prior(kernel=RBF(), mean_function=Constant())
    likelihood = Gaussian(num_datapoints=2)
    posterior = prior * likelihood

    with pytest.raises(scipy.optimize.OptimizeWarning):
        fit_scipy(
            model=posterior,
            objective=conjugate_mll,
            train_data=D,
            max_iters=1,
        )


@pytest.mark.parametrize("num_iters", [1, 5])
@pytest.mark.parametrize("batch_size", [1, 20, 50])
@pytest.mark.parametrize("n_data", [50])
@pytest.mark.parametrize("verbose", [True, False])
def test_fit_batch(num_iters: int, batch_size: int, n_data: int, verbose: bool) -> None:
    # Create dataset:
    key = jr.key(123)
    x = jnp.sort(
        jr.uniform(key=key, minval=-2.0, maxval=2.0, shape=(n_data, 1)), axis=0
    )
    y = jnp.sin(x) + jr.normal(key=key, shape=x.shape) * 0.1
    D = Dataset(X=x, y=y)

    # Define GP model:
    prior = Prior(kernel=RBF(), mean_function=Constant())
    likelihood = Gaussian(num_datapoints=n_data)
    posterior = prior * likelihood

    # Define variational family:
    z = jnp.linspace(-2.0, 2.0, 10).reshape(-1, 1)
    q = VariationalGaussian(posterior=posterior, inducing_inputs=z)

    # Train!
    trained_model, history = fit(
        model=q,
        objective=elbo,
        train_data=D,
        optim=ox.adam(0.1),
        num_iters=num_iters,
        batch_size=batch_size,
        verbose=verbose,
        key=jr.key(123),
    )

    # Ensure the trained model is a Gaussian process posterior
    assert isinstance(trained_model, VariationalGaussian)

    # Ensure we return a history of the correct length
    assert len(history) == num_iters

    # Ensure we reduce the loss
    assert elbo(trained_model, D) < elbo(q, D)


@pytest.mark.parametrize("n_data", [50])
@pytest.mark.parametrize("n_dim", [1, 2, 3])
@pytest.mark.parametrize("batch_size", [1, 2, 50])
def test_get_batch(n_data: int, n_dim: int, batch_size: int):
    key = jr.key(123)

    # Create dataset:
    x = jnp.sort(
        jr.uniform(key=key, minval=-2.0, maxval=2.0, shape=(n_data, n_dim)), axis=0
    )
    y = jnp.sin(x) + jr.normal(key=key, shape=x.shape) * 0.1
    D = Dataset(X=x, y=y)

    # Sample out a batch:
    B = get_batch(D, batch_size, key)

    # Check batch is correct size and shape dimensions:
    assert B.n == batch_size
    assert B.X.shape[1:] == x.shape[1:]
    assert B.y.shape[1:] == y.shape[1:]

    # Ensure no caching of batches:
    key, subkey = jr.split(key)
    New = get_batch(D, batch_size, subkey)
    assert New.n == batch_size
    assert New.X.shape[1:] == x.shape[1:]
    assert New.y.shape[1:] == y.shape[1:]
    assert jnp.sum(New.X == B.X) <= n_dim * batch_size / n_data
    assert jnp.sum(New.y == B.y) <= n_dim * batch_size / n_data


@pytest.fixture
def valid_model() -> nnx.Module:
    """Return a valid model for testing."""

    class LinearModel(nnx.Module):
        def __init__(self, weight: float, bias: float) -> None:
            self.weight = PositiveReal(weight)
            self.bias = bias

        def __call__(self, x: Any) -> Any:
            return self.weight[...] * x + self.bias

    return LinearModel(weight=1.0, bias=1.0)


@pytest.fixture
def valid_dataset() -> Dataset:
    """Return a valid dataset for testing."""
    X = jnp.array([[1.0], [2.0], [3.0]])
    y = jnp.array([[1.0], [2.0], [3.0]])
    return Dataset(X=X, y=y)


def test_check_model_valid(valid_model: nnx.Module) -> None:
    """Test that a valid model passes validation."""
    _check_model(valid_model)


def test_check_model_invalid() -> None:
    """Test that an invalid model raises a TypeError."""
    model = "not a model"
    with pytest.raises(
        TypeError, match=r"Expected model to be a subclass of nnx\.Module"
    ):
        _check_model(model)


def test_check_train_data_valid(valid_dataset: Dataset) -> None:
    """Test that valid training data passes validation."""
    _check_train_data(valid_dataset)


def test_check_train_data_invalid() -> None:
    """Test that invalid training data raises a TypeError."""
    train_data = "not a dataset"
    with pytest.raises(
        TypeError, match=r"Expected train_data to be of type gpjax\.Dataset"
    ):
        _check_train_data(train_data)


def test_check_optim_valid() -> None:
    """Test that a valid optimiser passes validation."""
    optim = ox.sgd(0.1)
    _check_optim(optim)


def test_check_optim_invalid() -> None:
    """Test that an invalid optimiser raises a TypeError."""
    optim = "not an optimiser"
    with pytest.raises(
        TypeError, match=r"Expected optim to be of type optax\.GradientTransformation"
    ):
        _check_optim(optim)


@pytest.mark.parametrize("num_iters", [1, 10, 100])
def test_check_num_iters_valid(num_iters: int) -> None:
    """Test that valid number of iterations passes validation."""
    _check_num_iters(num_iters)


def test_check_num_iters_invalid_type() -> None:
    """Test that an invalid num_iters type raises a TypeError."""
    num_iters = "not an int"
    with pytest.raises(TypeError, match="Expected num_iters to be of type int"):
        _check_num_iters(num_iters)


@pytest.mark.parametrize("num_iters", [0, -5])
def test_check_num_iters_invalid_value(num_iters: int) -> None:
    """Test that an invalid num_iters value raises a ValueError."""
    with pytest.raises(ValueError, match="Expected num_iters to be positive"):
        _check_num_iters(num_iters)


@pytest.mark.parametrize("log_rate", [1, 10, 100])
def test_check_log_rate_valid(log_rate: int) -> None:
    """Test that a valid log rate passes validation."""
    _check_log_rate(log_rate)


def test_check_log_rate_invalid_type() -> None:
    """Test that an invalid log_rate type raises a TypeError."""
    log_rate = "not an int"
    with pytest.raises(TypeError, match="Expected log_rate to be of type int"):
        _check_log_rate(log_rate)


@pytest.mark.parametrize("log_rate", [0, -5])
def test_check_log_rate_invalid_value(log_rate: int) -> None:
    """Test that an invalid log_rate value raises a ValueError."""
    with pytest.raises(ValueError, match="Expected log_rate to be positive"):
        _check_log_rate(log_rate)


@pytest.mark.parametrize("verbose", [True, False])
def test_check_verbose_valid(verbose: bool) -> None:
    """Test that valid verbose values pass validation."""
    _check_verbose(verbose)


def test_check_verbose_invalid() -> None:
    """Test that an invalid verbose value raises a TypeError."""
    verbose = "not a bool"
    with pytest.raises(TypeError, match="Expected verbose to be of type bool"):
        _check_verbose(verbose)


@pytest.mark.parametrize("batch_size", [1, 10, 100, -1])
def test_check_batch_size_valid(batch_size: int) -> None:
    """Test that valid batch sizes pass validation."""
    _check_batch_size(batch_size)


def test_check_batch_size_invalid_type() -> None:
    """Test that an invalid batch_size type raises a TypeError."""
    batch_size = "not an int"
    with pytest.raises(TypeError, match="Expected batch_size to be of type int"):
        _check_batch_size(batch_size)


@pytest.mark.parametrize("batch_size", [0, -2, -5])
def test_check_batch_size_invalid_value(batch_size: int) -> None:
    """Test that invalid batch_size values raise a ValueError."""
    with pytest.raises(ValueError, match="Expected batch_size to be positive or -1"):
        _check_batch_size(batch_size)


def test_fit_filter_freeze_kernel_variance() -> None:
    """Test that fit can freeze kernel variance parameter using filters."""
    key = jr.key(42)
    X = jr.uniform(key, (20, 1), minval=-3.0, maxval=3.0)
    y = jnp.sin(X) + 0.1 * jr.normal(jr.key(43), (20, 1))
    D = Dataset(X, y)

    # Create GP with RBF kernel
    meanf = gpx.mean_functions.Zero()
    kernel = gpx.kernels.RBF(lengthscale=1.0, variance=1.0)
    prior = gpx.gps.Prior(mean_function=meanf, kernel=kernel)
    likelihood = gpx.likelihoods.Gaussian(num_datapoints=D.n)
    posterior = prior * likelihood

    # Record initial variance value
    initial_variance = kernel.variance[...]

    # Train with filter that excludes variance (freezes it)
    filter_no_variance = nnx.filterlib.Not(nnx.filterlib.PathContains("variance"))
    trained_posterior, _ = fit(
        model=posterior,
        objective=gpx.objectives.conjugate_mll,
        train_data=D,
        trainable=filter_no_variance,
        optim=ox.sgd(0.01),
        num_iters=10,
        verbose=False,
    )

    # Assert variance has not changed
    assert jnp.allclose(trained_posterior.prior.kernel.variance[...], initial_variance)

    # Assert lengthscale has changed
    assert not jnp.allclose(trained_posterior.prior.kernel.lengthscale[...], 1.0)


def test_fit_zero_mean_function_not_trained() -> None:
    """Test that Zero mean function constant is not trained even with default filter."""
    key = jr.key(42)
    X = jr.uniform(key, (20, 1), minval=-3.0, maxval=3.0)
    y = jnp.ones_like(X) + 0.1 * jr.normal(jr.key(43), X.shape)  # Non-zero mean data
    D = Dataset(X, y)

    # Create GP with Zero mean function
    meanf = gpx.mean_functions.Zero()
    kernel = gpx.kernels.RBF(lengthscale=1.0, variance=1.0)
    prior = gpx.gps.Prior(mean_function=meanf, kernel=kernel)
    likelihood = gpx.likelihoods.Gaussian(num_datapoints=D.n)
    posterior = prior * likelihood

    # Record initial mean function constant (should be 0.0)
    initial_constant = meanf.constant

    # Train with default filter (should not train Zero mean function's constant)
    trained_posterior, _ = fit(
        model=posterior,
        objective=gpx.objectives.conjugate_mll,
        train_data=D,
        optim=ox.sgd(0.01),
        num_iters=10,
        verbose=False,
    )

    # Assert Zero mean function constant has not changed (remains 0.0)
    assert jnp.allclose(
        trained_posterior.prior.mean_function.constant, initial_constant
    )


def test_fit_constant_mean_function_with_parameter() -> None:
    """Test that Constant mean function works with trainable Parameter."""
    key = jr.key(42)
    X = jr.uniform(key, (20, 1), minval=-3.0, maxval=3.0)
    y = 5.0 * jnp.ones_like(X) + 0.1 * jr.normal(jr.key(43), X.shape)  # Mean of 5.0
    D = Dataset(X, y)

    # Create GP with Constant mean function using Parameter
    from gpjax.parameters import Real

    meanf = gpx.mean_functions.Constant(constant=Real(1.0))  # Start with mean 1.0
    kernel = gpx.kernels.RBF(lengthscale=1.0, variance=1.0)
    prior = gpx.gps.Prior(mean_function=meanf, kernel=kernel)
    likelihood = gpx.likelihoods.Gaussian(num_datapoints=D.n, obs_stddev=0.1)
    posterior = prior * likelihood

    # Record initial mean function constant
    initial_constant = meanf.constant[...]

    # Train with default filter (should train the mean function Parameter)
    trained_posterior, _ = fit(
        model=posterior,
        objective=gpx.objectives.conjugate_mll,
        train_data=D,
        optim=ox.adam(0.01),
        num_iters=20,
        verbose=False,
    )

    # Assert mean function constant has changed (parameter is trainable)
    final_constant = trained_posterior.prior.mean_function.constant[...]
    assert not jnp.allclose(final_constant, initial_constant)
    # Just verify the parameter changed (direction depends on optimization dynamics)
    assert jnp.isfinite(final_constant)  # Not NaN/Inf


def test_fit_constant_mean_function_with_raw_value() -> None:
    """Test that Constant mean function works with fixed raw value."""
    key = jr.key(42)
    X = jr.uniform(key, (20, 1), minval=-3.0, maxval=3.0)
    y = 5.0 * jnp.ones_like(X) + 0.1 * jr.normal(jr.key(43), X.shape)  # Mean of 5.0
    D = Dataset(X, y)

    # Create GP with Constant mean function using raw value
    meanf = gpx.mean_functions.Constant(constant=1.0)  # Fixed mean 1.0
    kernel = gpx.kernels.RBF(lengthscale=1.0, variance=0.1)
    prior = gpx.gps.Prior(mean_function=meanf, kernel=kernel)
    likelihood = gpx.likelihoods.Gaussian(num_datapoints=D.n, obs_stddev=0.1)
    posterior = prior * likelihood

    # Record initial mean function constant
    initial_constant = meanf.constant

    # Train with default filter (should NOT train the raw value)
    trained_posterior, _ = fit(
        model=posterior,
        objective=gpx.objectives.conjugate_mll,
        train_data=D,
        optim=ox.sgd(0.1),
        num_iters=50,
        verbose=False,
    )

    # Assert mean function constant has NOT changed (fixed raw value)
    final_constant = trained_posterior.prior.mean_function.constant
    assert jnp.allclose(final_constant, initial_constant)


def test_fit_filter_by_type() -> None:
    """Test filtering parameters by type using nnx.filters.OfType."""
    key = jr.key(42)
    X = jr.uniform(key, (20, 1), minval=-3.0, maxval=3.0)
    y = jnp.sin(X) + 0.1 * jr.normal(jr.key(43), (20, 1))
    D = Dataset(X, y)

    # Create GP with RBF kernel
    from gpjax.parameters import PositiveReal

    meanf = gpx.mean_functions.Zero()
    kernel = gpx.kernels.RBF(lengthscale=1.0, variance=1.0)
    prior = gpx.gps.Prior(mean_function=meanf, kernel=kernel)
    likelihood = gpx.likelihoods.Gaussian(num_datapoints=D.n)
    posterior = prior * likelihood

    # Record initial values
    initial_variance = kernel.variance[...]
    initial_lengthscale = kernel.lengthscale[...]
    initial_obs_stddev = likelihood.obs_stddev[...]

    # Train only PositiveReal parameters (should include only lengthscale)
    filter_positive_real = nnx.filterlib.OfType(PositiveReal)
    trained_posterior, _ = fit(
        model=posterior,
        objective=gpx.objectives.conjugate_mll,
        train_data=D,
        trainable=filter_positive_real,
        optim=ox.sgd(0.01),
        num_iters=10,
        verbose=False,
    )

    # Assert that only PositiveReal parameters (lengthscale) have changed
    # variance and obs_stddev are NonNegativeReal, so they should not change
    assert jnp.allclose(trained_posterior.prior.kernel.variance[...], initial_variance)
    assert not jnp.allclose(
        trained_posterior.prior.kernel.lengthscale[...], initial_lengthscale
    )
    assert jnp.allclose(
        trained_posterior.likelihood.obs_stddev[...], initial_obs_stddev
    )



"""Linear operator abstractions for GPJax."""

from abc import (
    ABC,
    abstractmethod,
)
from typing import (
    Any,
)

from jax import Array
import jax.numpy as jnp
import jax.tree_util as jtu
from jaxtyping import Float


class LinearOperator(ABC):
    """Abstract base class for linear operators."""

    def __init__(self):
        super().__init__()

    @property
    @abstractmethod
    def shape(self) -> tuple[int, int]:
        """Return the shape of the operator."""

    @property
    @abstractmethod
    def dtype(self) -> jnp.dtype:
        """Return the data type of the operator."""

    @abstractmethod
    def to_dense(self) -> Float[Array, "M N"]:
        """Convert the operator to a dense JAX array."""

    @property
    def T(self) -> "LinearOperator":
        """Return the transpose of the operator."""
        # Default implementation: convert to dense and transpose
        return Dense(self.to_dense().T)

    def __matmul__(self, other):
        """Matrix multiplication with another array or operator."""
        if hasattr(other, "to_dense"):
            # Other is a LinearOperator
            return Dense(self.to_dense() @ other.to_dense())
        else:
            # Other is a JAX array
            return self.to_dense() @ other

    def __rmatmul__(self, other):
        """Right matrix multiplication (other @ self)."""
        if hasattr(other, "to_dense"):
            # Other is a LinearOperator
            return Dense(other.to_dense() @ self.to_dense())
        else:
            # Other is a JAX array
            return other @ self.to_dense()

    def __add__(self, other):
        """Addition with another array or operator."""
        if hasattr(other, "to_dense"):
            # Other is a LinearOperator
            return Dense(self.to_dense() + other.to_dense())
        else:
            # Other is a JAX array
            return Dense(self.to_dense() + other)

    def __radd__(self, other):
        """Right addition (other + self)."""
        if hasattr(other, "to_dense"):
            # Other is a LinearOperator
            return Dense(other.to_dense() + self.to_dense())
        else:
            # Other is a JAX array
            return Dense(other + self.to_dense())

    def __sub__(self, other):
        """Subtraction with another array or operator."""
        if hasattr(other, "to_dense"):
            # Other is a LinearOperator
            return Dense(self.to_dense() - other.to_dense())
        else:
            # Other is a JAX array
            return Dense(self.to_dense() - other)

    def __rsub__(self, other):
        """Right subtraction (other - self)."""
        if hasattr(other, "to_dense"):
            # Other is a LinearOperator
            return Dense(other.to_dense() - self.to_dense())
        else:
            # Other is a JAX array
            return Dense(other - self.to_dense())

    def __mul__(self, other):
        """Scalar multiplication (self * scalar)."""
        if jnp.isscalar(other):
            return Dense(self.to_dense() * other)
        else:
            # Element-wise multiplication with array
            return Dense(self.to_dense() * other)

    def __rmul__(self, other):
        """Right scalar multiplication (scalar * self)."""
        if jnp.isscalar(other):
            return Dense(other * self.to_dense())
        else:
            # Element-wise multiplication with array
            return Dense(other * self.to_dense())


class Dense(LinearOperator):
    """Dense linear operator wrapping a JAX array."""

    def __init__(self, array: Float[Array, "M N"]):
        super().__init__()
        self.array = array

    @property
    def shape(self) -> tuple[int, int]:
        return self.array.shape

    @property
    def dtype(self) -> jnp.dtype:
        return self.array.dtype

    def to_dense(self) -> Float[Array, "M N"]:
        return self.array

    @property
    def T(self) -> "Dense":
        return Dense(self.array.T)


class Diagonal(LinearOperator):
    """Diagonal linear operator."""

    def __init__(self, diagonal: Float[Array, " N"]):
        super().__init__()
        self.diagonal = diagonal

    @property
    def shape(self) -> tuple[int, int]:
        n = self.diagonal.shape[0]
        return (n, n)

    @property
    def dtype(self) -> jnp.dtype:
        return self.diagonal.dtype

    def to_dense(self) -> Float[Array, "N N"]:
        return jnp.diag(self.diagonal)

    @property
    def T(self) -> "Diagonal":
        return Diagonal(self.diagonal)


class Identity(LinearOperator):
    """Identity linear operator."""

    def __init__(self, shape: int | tuple[int, int], dtype=jnp.float64):
        super().__init__()
        if isinstance(shape, int):
            self._shape = (shape, shape)
        else:
            if shape[0] != shape[1]:
                raise ValueError(f"Identity matrix must be square, got shape {shape}")
            self._shape = shape
        self._dtype = dtype

    @property
    def shape(self) -> tuple[int, int]:
        return self._shape

    @property
    def dtype(self) -> Any:
        return self._dtype

    def to_dense(self) -> Float[Array, "N N"]:
        n = self._shape[0]
        return jnp.eye(n, dtype=self._dtype)

    @property
    def T(self) -> "Identity":
        return Identity(self._shape, dtype=self._dtype)


class Triangular(LinearOperator):
    """Triangular linear operator."""

    def __init__(self, array: Float[Array, "N N"], lower: bool = True):
        super().__init__()
        self.array = array
        self.lower = lower

    @property
    def shape(self) -> tuple[int, int]:
        return self.array.shape

    @property
    def dtype(self) -> Any:
        return self.array.dtype

    def to_dense(self) -> Float[Array, "N N"]:
        if self.lower:
            return jnp.tril(self.array)
        else:
            return jnp.triu(self.array)

    @property
    def T(self) -> "Triangular":
        return Triangular(self.array.T, lower=not self.lower)


class BlockDiag(LinearOperator):
    """Block diagonal linear operator."""

    def __init__(
        self, operators: list[LinearOperator], multiplicities: list[int] | None = None
    ):
        super().__init__()
        self.operators = operators

        # Handle multiplicities - how many times each block is repeated
        if multiplicities is None:
            self.multiplicities = [1] * len(operators)
        else:
            if len(multiplicities) != len(operators):
                raise ValueError(
                    f"Length of multiplicities ({len(multiplicities)}) must match operators ({len(operators)})"
                )
            self.multiplicities = multiplicities

        # Calculate total shape with multiplicities
        rows = sum(
            op.shape[0] * mult
            for op, mult in zip(operators, self.multiplicities, strict=False)
        )
        cols = sum(
            op.shape[1] * mult
            for op, mult in zip(operators, self.multiplicities, strict=False)
        )
        self._shape = (rows, cols)

        # Use dtype of first operator (assuming all same dtype)
        if operators:
            self._dtype = operators[0].dtype
        else:
            self._dtype = jnp.float64

    @property
    def shape(self) -> tuple[int, int]:
        return self._shape

    @property
    def dtype(self) -> Any:
        return self._dtype

    def to_dense(self) -> Float[Array, "M N"]:
        if not self.operators:
            return jnp.zeros(self._shape, dtype=self._dtype)

        # Convert each operator to dense and create block diagonal with multiplicities
        expanded_blocks = []
        for op, mult in zip(self.operators, self.multiplicities, strict=False):
            op_dense = op.to_dense()
            for _ in range(mult):
                expanded_blocks.append(op_dense)

        # Create the full block diagonal matrix
        n_blocks = len(expanded_blocks)
        if n_blocks == 0:
            return jnp.zeros(self._shape, dtype=self._dtype)

        # Build the block diagonal matrix
        rows = []
        for i in range(n_blocks):
            row = []
            for j in range(n_blocks):
                if i == j:
                    row.append(expanded_blocks[i])
                else:
                    row.append(
                        jnp.zeros(
                            (expanded_blocks[i].shape[0], expanded_blocks[j].shape[1]),
                            dtype=self._dtype,
                        )
                    )
            rows.append(row)
        return jnp.block(rows)

    @property
    def T(self) -> "BlockDiag":
        transposed_ops = [op.T for op in self.operators]
        return BlockDiag(transposed_ops, multiplicities=self.multiplicities)


class Kronecker(LinearOperator):
    """Kronecker product linear operator."""

    def __init__(self, operators: list[LinearOperator]):
        super().__init__()
        if len(operators) < 2:
            raise ValueError("Kronecker product requires at least 2 operators")
        self.operators = operators

        # Calculate shape as product of individual shapes
        rows = 1
        cols = 1
        for op in operators:
            rows *= op.shape[0]
            cols *= op.shape[1]
        self._shape = (rows, cols)

        # Use dtype of first operator
        self._dtype = operators[0].dtype

    @property
    def shape(self) -> tuple[int, int]:
        return self._shape

    @property
    def dtype(self) -> Any:
        return self._dtype

    def to_dense(self) -> Float[Array, "M N"]:
        # Convert to dense and compute Kronecker product
        result = self.operators[0].to_dense()
        for op in self.operators[1:]:
            result = jnp.kron(result, op.to_dense())
        return result

    @property
    def T(self) -> "Kronecker":
        transposed_ops = [op.T for op in self.operators]
        return Kronecker(transposed_ops)


def _dense_tree_flatten(dense):
    return (dense.array,), None


def _dense_tree_unflatten(aux_data, children):
    return Dense(children[0])


jtu.register_pytree_node(Dense, _dense_tree_flatten, _dense_tree_unflatten)


def _diagonal_tree_flatten(diagonal):
    return (diagonal.diagonal,), None


def _diagonal_tree_unflatten(aux_data, children):
    return Diagonal(children[0])


jtu.register_pytree_node(Diagonal, _diagonal_tree_flatten, _diagonal_tree_unflatten)


def _identity_tree_flatten(identity):
    return (), (identity._shape, identity._dtype)


def _identity_tree_unflatten(aux_data, children):
    shape, dtype = aux_data
    return Identity(shape, dtype)


jtu.register_pytree_node(Identity, _identity_tree_flatten, _identity_tree_unflatten)


def _triangular_tree_flatten(triangular):
    return (triangular.array,), triangular.lower


def _triangular_tree_unflatten(aux_data, children):
    return Triangular(children[0], aux_data)


jtu.register_pytree_node(
    Triangular, _triangular_tree_flatten, _triangular_tree_unflatten
)


def _blockdiag_tree_flatten(blockdiag):
    return tuple(blockdiag.operators), blockdiag.multiplicities


def _blockdiag_tree_unflatten(aux_data, children):
    return BlockDiag(list(children), aux_data)


jtu.register_pytree_node(BlockDiag, _blockdiag_tree_flatten, _blockdiag_tree_unflatten)


def _kronecker_tree_flatten(kronecker):
    return tuple(kronecker.operators), None


def _kronecker_tree_unflatten(aux_data, children):
    return Kronecker(list(children))


jtu.register_pytree_node(Kronecker, _kronecker_tree_flatten, _kronecker_tree_unflatten)

__all__ = [
    "BlockDiag",
    "Dense",
    "Diagonal",
    "Identity",
    "Kronecker",
    "LinearOperator",
    "Triangular",
]



"""Linear algebra module for GPJax."""

from gpjax.linalg.operations import (
    diag,
    logdet,
    lower_cholesky,
    solve,
)
from gpjax.linalg.operators import (
    BlockDiag,
    Dense,
    Diagonal,
    Identity,
    Kronecker,
    LinearOperator,
    Triangular,
)
from gpjax.linalg.utils import (
    PSD,
    psd,
)

__all__ = [
    "PSD",
    "BlockDiag",
    "Dense",
    "Diagonal",
    "Identity",
    "Kronecker",
    "LinearOperator",
    "Triangular",
    "diag",
    "logdet",
    "lower_cholesky",
    "psd",
    "solve",
]



"""Utility functions for the linear algebra module."""

import jax.numpy as jnp
from jaxtyping import Array

from gpjax.linalg.operators import LinearOperator


class PSDAnnotation:
    """Marker class for PSD (Positive Semi-Definite) annotations."""

    def __call__(self, A: LinearOperator) -> LinearOperator:
        """Make PSD annotation callable."""
        return psd(A)


# Create the PSD marker similar to cola.PSD
PSD = PSDAnnotation()


def psd(A: LinearOperator) -> LinearOperator:
    """Mark a linear operator as positive semi-definite.

    This function acts as a marker/wrapper for positive semi-definite matrices.

    Args:
        A: A LinearOperator that is assumed to be positive semi-definite.

    Returns:
        The same LinearOperator, marked as PSD.
    """
    # Add annotations attribute if it doesn't exist
    if not hasattr(A, "annotations"):
        A.annotations = set()
    A.annotations.add(PSD)
    return A


def add_jitter(matrix: Array, jitter: float | Array = 1e-6) -> Array:
    """Add jitter to the diagonal of a matrix for numerical stability.

    This function adds a small positive value (jitter) to the diagonal elements
    of a square matrix to improve numerical stability, particularly for
    Cholesky decompositions and matrix inversions.

    Args:
        matrix: A square matrix to which jitter will be added.
        jitter: The jitter value to add to the diagonal. Defaults to 1e-6.

    Returns:
        The matrix with jitter added to its diagonal.

    Examples:
        >>> import jax.numpy as jnp
        >>> from gpjax.linalg.utils import add_jitter
        >>> matrix = jnp.array([[1.0, 0.5], [0.5, 1.0]])
        >>> jittered_matrix = add_jitter(matrix, jitter=0.01)
    """
    if matrix.ndim != 2:
        raise ValueError(f"Expected 2D matrix, got {matrix.ndim}D array")

    if matrix.shape[0] != matrix.shape[1]:
        raise ValueError(f"Expected square matrix, got shape {matrix.shape}")

    return matrix + jnp.eye(matrix.shape[0]) * jitter



# # Copyright 2022 The thomaspinder Contributors. All Rights Reserved.
# #
# # you may not use this file except in compliance with the License.
# # You may obtain a copy of the License at
# #
# #
# # Unless required by applicable law or agreed to in writing, software
# # distributed under the License is distributed on an "AS IS" BASIS,
# # WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# # See the License for the specific language governing permissions and
# # limitations under the License.

from gpjax.kernels.non_euclidean import GraphKernel
from gpjax.kernels.non_euclidean.utils import (
    calculate_heat_semigroup,
    jax_gather_nd,
)
from gpjax.linalg.operators import Identity
from jax import (
    config,
    jit,
    vmap,
)
import jax.numpy as jnp
import networkx as nx
import pytest

# # Enable Float64 for more stable matrix inversions.
config.update("jax_enable_x64", True)


def test_graph_kernel():
    # Create a random graph, G, and verice labels, x,
    n_verticies = 20
    n_edges = 40
    G = nx.gnm_random_graph(n_verticies, n_edges, seed=123)
    x = jnp.arange(n_verticies).reshape(-1, 1)

    # Compute graph laplacian
    L = nx.laplacian_matrix(G).toarray() + jnp.eye(n_verticies) * 1e-12

    # Create graph kernel
    kern = GraphKernel(
        laplacian=L,
    )
    assert isinstance(kern, GraphKernel)
    assert kern.num_vertex == n_verticies
    assert kern.eigenvalues.shape == (n_verticies, 1)
    assert kern.eigenvectors.shape == (n_verticies, n_verticies)

    # Compute gram matrix
    Kxx = kern.gram(x)
    assert Kxx.shape == (n_verticies, n_verticies)

    # Check positive definiteness
    Kxx += Identity(Kxx.shape[0]) * 1e-6
    eigen_values = jnp.linalg.eigvalsh(Kxx.to_dense())
    assert all(eigen_values > 0)


def _build_test_kernel(n_vertices: int = 10) -> GraphKernel:
    graph = nx.path_graph(n_vertices)
    laplacian = nx.laplacian_matrix(graph).toarray() + jnp.eye(n_vertices) * 1e-12
    return GraphKernel(laplacian=laplacian)


def test_graph_kernel_accepts_vector_indices():
    kernel = _build_test_kernel()
    col_indices = jnp.arange(5).reshape(-1, 1)
    vector_indices = col_indices.squeeze()

    matrix_eval = kernel(col_indices, col_indices)
    vector_eval = kernel(vector_indices, vector_indices)

    assert matrix_eval.shape == (5, 5)
    assert vector_eval.shape == (5, 5)
    assert jnp.allclose(matrix_eval, vector_eval)


def test_graph_kernel_vmappable_over_single_indices():
    kernel = _build_test_kernel()
    idx = jnp.arange(4)

    diag_entries = vmap(lambda z: kernel(z, z))(idx)
    assert diag_entries.shape == (4,)
    assert jnp.all(diag_entries >= 0.0)


def test_graph_kernel_vmappable_over_pairs():
    kernel = _build_test_kernel()
    x = jnp.arange(5)
    y = jnp.array([4, 3, 2, 1, 0])

    vectorised_eval = vmap(lambda a, b: kernel(a, b))(x, y)
    baseline = jnp.asarray([kernel(a, b) for a, b in zip(x, y, strict=False)])

    assert vectorised_eval.shape == (5,)
    assert jnp.allclose(vectorised_eval, baseline)


def test_graph_kernel_is_jittable():
    kernel = _build_test_kernel()
    jit_kernel = jit(lambda a, b: kernel(a, b))

    column = jnp.arange(5).reshape(-1, 1)
    vector = jnp.arange(5)
    pairs = [
        (0, 0),
        (0, column),
        (vector, 1),
        (vector, vector),
        (column, column),
    ]

    for x, y in pairs:
        expected = kernel(x, y)
        result = jit_kernel(x, y)
        assert result.shape == expected.shape
        assert jnp.allclose(result, expected)


# ---------------------------------------------------------------------------
# jax_gather_nd utility
# ---------------------------------------------------------------------------


class TestJaxGatherNd:
    """Tests for jax_gather_nd utility function."""

    def test_basic_1d_gather(self):
        params = jnp.array([10.0, 20.0, 30.0, 40.0, 50.0])
        indices = jnp.array([[0], [2], [4]])
        result = jax_gather_nd(params, indices)
        assert jnp.allclose(result, jnp.array([10.0, 30.0, 50.0]))

    def test_2d_gather_rows(self):
        params = jnp.array([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]])
        indices = jnp.array([[0], [2]])
        result = jax_gather_nd(params, indices)
        expected = jnp.array([[1.0, 2.0], [5.0, 6.0]])
        assert jnp.allclose(result, expected)

    def test_single_index(self):
        params = jnp.arange(10.0)
        indices = jnp.array([[3]])
        result = jax_gather_nd(params, indices)
        assert jnp.allclose(result, jnp.array([3.0]))

    def test_all_indices(self):
        params = jnp.array([100.0, 200.0, 300.0])
        indices = jnp.arange(3).reshape(-1, 1)
        result = jax_gather_nd(params, indices)
        assert jnp.allclose(result, params)

    def test_repeated_indices(self):
        params = jnp.array([10.0, 20.0, 30.0])
        indices = jnp.array([[1], [1], [1]])
        result = jax_gather_nd(params, indices)
        assert jnp.allclose(result, jnp.array([20.0, 20.0, 20.0]))


# ---------------------------------------------------------------------------
# calculate_heat_semigroup utility
# ---------------------------------------------------------------------------


class TestCalculateHeatSemigroup:
    """Tests for calculate_heat_semigroup utility function."""

    @pytest.fixture()
    def graph_kernel(self):
        n_vertices = 10
        graph = nx.path_graph(n_vertices)
        laplacian = nx.laplacian_matrix(graph).toarray() + jnp.eye(n_vertices) * 1e-12
        return GraphKernel(laplacian=laplacian)

    def test_output_shape(self, graph_kernel):
        S = calculate_heat_semigroup(graph_kernel)
        assert S.shape == (graph_kernel.num_vertex, 1)

    def test_finite_values(self, graph_kernel):
        S = calculate_heat_semigroup(graph_kernel)
        assert jnp.all(jnp.isfinite(S))

    def test_positive_values(self, graph_kernel):
        S = calculate_heat_semigroup(graph_kernel)
        assert jnp.all(S > 0.0)

    def test_scales_with_variance(self):
        """Doubling kernel variance should double the semigroup values."""
        n_vertices = 8
        graph = nx.path_graph(n_vertices)
        laplacian = nx.laplacian_matrix(graph).toarray() + jnp.eye(n_vertices) * 1e-12

        k1 = GraphKernel(laplacian=laplacian, variance=1.0)
        k2 = GraphKernel(laplacian=laplacian, variance=2.0)

        S1 = calculate_heat_semigroup(k1)
        S2 = calculate_heat_semigroup(k2)
        assert jnp.allclose(S2, 2.0 * S1, atol=1e-6)

    def test_sums_to_num_vertex_times_variance(self, graph_kernel):
        """The sum of S should equal num_vertex * variance (by construction)."""
        S = calculate_heat_semigroup(graph_kernel)
        expected_sum = graph_kernel.num_vertex * graph_kernel.variance[...]
        assert jnp.allclose(jnp.sum(S), expected_sum, atol=1e-5)



"""Tests for GPJax imports and exports."""

import gpjax


def test_all_exports_importable():
    """Test that all items in __all__ are actually importable."""
    for item in gpjax.__all__:
        assert hasattr(gpjax, item), f"{item} not available in gpjax module"


def test_common_imports():
    """Test that common import patterns work correctly."""
    # Test basic imports
    assert hasattr(gpjax, "gps")
    assert hasattr(gpjax, "kernels")
    assert hasattr(gpjax, "likelihoods")
    assert hasattr(gpjax, "Dataset")
    assert hasattr(gpjax, "fit")

    # Test callable functions
    assert callable(gpjax.fit)
    assert callable(gpjax.fit_lbfgs)
    assert callable(gpjax.fit_scipy)
    assert callable(gpjax.cite)



# Copyright 2024 The GPJax Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""JIT compatibility smoke tests.

Verifies that core GPJax operations (kernel evaluation, GP prediction,
linalg operators) produce identical results with and without
``jax.jit``.
"""

from gpjax.dataset import Dataset
from gpjax.gps import Prior
from gpjax.kernels.stationary import RBF, Matern32
from gpjax.likelihoods import Gaussian
from gpjax.linalg import Dense, Diagonal, Identity
from gpjax.mean_functions import Constant
import jax
from jax import config
import jax.numpy as jnp
import pytest

config.update("jax_enable_x64", True)

# Shared fixtures ----------------------------------------------------------- #

N_TRAIN = 20
N_TEST = 5
D = 1


@pytest.fixture
def train_data():
    x = jnp.linspace(0, 1, N_TRAIN)[:, None]
    y = jnp.sin(x)
    return Dataset(X=x, y=y)


@pytest.fixture
def test_inputs():
    return jnp.linspace(0, 1, N_TEST)[:, None]


# --------------------------------------------------------------------------- #
#  Kernel JIT tests                                                            #
# --------------------------------------------------------------------------- #

KERNELS = [RBF, Matern32]


@pytest.mark.parametrize("KernelClass", KERNELS)
def test_kernel_gram_jit(KernelClass):
    x = jnp.linspace(0, 1, N_TRAIN)[:, None]
    kernel = KernelClass()

    def gram_fn(x):
        return kernel.gram(x).to_dense()

    result = gram_fn(x)
    result_jit = jax.jit(gram_fn)(x)
    assert jnp.allclose(result, result_jit, atol=1e-12)


@pytest.mark.parametrize("KernelClass", KERNELS)
def test_kernel_cross_covariance_jit(KernelClass):
    x = jnp.linspace(0, 1, N_TRAIN)[:, None]
    y = jnp.linspace(0.5, 1.5, N_TEST)[:, None]
    kernel = KernelClass()

    def cross_cov_fn(x, y):
        return kernel.cross_covariance(x, y)

    result = cross_cov_fn(x, y)
    result_jit = jax.jit(cross_cov_fn)(x, y)
    assert jnp.allclose(result, result_jit, atol=1e-12)


# --------------------------------------------------------------------------- #
#  GP Prior JIT tests                                                          #
# --------------------------------------------------------------------------- #


def test_prior_predict_jit(test_inputs):
    prior = Prior(mean_function=Constant(), kernel=RBF())

    def predict_fn(x):
        dist = prior.predict(x)
        return dist.mean, dist.covariance()

    mean, cov = predict_fn(test_inputs)
    mean_jit, cov_jit = jax.jit(predict_fn)(test_inputs)

    assert jnp.allclose(mean, mean_jit, atol=1e-12)
    assert jnp.allclose(cov, cov_jit, atol=1e-12)


# --------------------------------------------------------------------------- #
#  GP Posterior JIT tests                                                      #
# --------------------------------------------------------------------------- #


def test_conjugate_posterior_predict_jit(train_data, test_inputs):
    prior = Prior(mean_function=Constant(), kernel=RBF())
    likelihood = Gaussian(num_datapoints=train_data.n)
    posterior = prior * likelihood

    def predict_fn(x_test, data):
        dist = posterior.predict(x_test, data)
        return dist.mean, dist.covariance()

    mean, cov = predict_fn(test_inputs, train_data)
    mean_jit, cov_jit = jax.jit(predict_fn)(test_inputs, train_data)

    assert jnp.allclose(mean, mean_jit, atol=1e-12)
    assert jnp.allclose(cov, cov_jit, atol=1e-12)


# --------------------------------------------------------------------------- #
#  Linalg operator JIT tests                                                   #
# --------------------------------------------------------------------------- #


def test_dense_matmul_jit():
    A = Dense(jnp.eye(3))
    v = jnp.ones((3, 1))

    def matmul_fn(v):
        return A @ v

    result = matmul_fn(v)
    result_jit = jax.jit(matmul_fn)(v)
    assert jnp.allclose(result, result_jit, atol=1e-12)


def test_diagonal_to_dense_jit():
    d = jnp.array([1.0, 2.0, 3.0])

    def to_dense_fn(d):
        return Diagonal(d).to_dense()

    result = to_dense_fn(d)
    result_jit = jax.jit(to_dense_fn)(d)
    assert jnp.allclose(result, result_jit, atol=1e-12)


def test_identity_to_dense_jit():
    def to_dense_fn():
        return Identity((3, 3)).to_dense()

    result = to_dense_fn()
    result_jit = jax.jit(to_dense_fn)()
    assert jnp.allclose(result, result_jit, atol=1e-12)



# Copyright 2022 The GPJax Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


from beartype.typing import (
    Optional,
)
from jax import vmap
import jax.numpy as jnp
import jax.random as jr
from jaxtyping import Float
from numpyro.distributions import constraints
from numpyro.distributions.distribution import Distribution
from numpyro.distributions.util import is_prng_key

from gpjax.linalg.operations import (
    diag,
    logdet,
    lower_cholesky,
    solve,
)
from gpjax.linalg.operators import LinearOperator
from gpjax.linalg.utils import psd
from gpjax.typing import (
    Array,
    ScalarFloat,
)


class GaussianDistribution(Distribution):
    r"""Multivariate Gaussian distribution for GP predictions.

    This is the return type of all ``predict()`` methods in GPJax. It wraps a
    mean vector and a covariance :class:`~gpjax.linalg.operators.LinearOperator`,
    providing methods for sampling, computing log-probabilities, and evaluating
    KL divergences.

    The distribution is parameterised as

    .. math::

        p(\mathbf{x}) = \mathcal{N}(\mathbf{x}; \boldsymbol{\mu}, \mathbf{\Sigma})

    where :math:`\boldsymbol{\mu}` is the ``loc`` (mean) vector and
    :math:`\mathbf{\Sigma}` is represented by the ``scale``
    :class:`~gpjax.linalg.operators.LinearOperator`. The ``scale`` is
    automatically annotated as positive semi-definite on construction.

    Parameters
    ----------
    loc : Float[Array, " N"]
        Mean vector of the distribution.
    scale : LinearOperator
        Covariance matrix represented as a
        :class:`~gpjax.linalg.operators.LinearOperator` (e.g.
        :class:`~gpjax.linalg.operators.Dense` or
        :class:`~gpjax.linalg.operators.Diagonal`).

    Examples
    --------
    >>> import jax.numpy as jnp
    >>> from gpjax.distributions import GaussianDistribution
    >>> from gpjax.linalg.operators import Dense
    >>> mu = jnp.array([0.0, 1.0])
    >>> cov = Dense(jnp.eye(2))
    >>> dist = GaussianDistribution(loc=mu, scale=cov)
    >>> dist.mean
    Array([0., 1.], dtype=float32)
    >>> dist.variance
    Array([1., 1.], dtype=float32)
    """

    support = constraints.real_vector

    def __init__(
        self,
        loc: Optional[Float[Array, " N"]],
        scale: Optional[LinearOperator],
        validate_args=None,
    ):
        self.loc = loc
        self.scale = psd(scale)
        batch_shape = ()
        event_shape = jnp.shape(self.loc)
        super().__init__(batch_shape, event_shape, validate_args=validate_args)

    def sample(self, key, sample_shape=()):
        r"""Draw samples from the distribution.

        Generates samples via the reparameterisation trick:

        .. math::

            \mathbf{x} = \boldsymbol{\mu} + \mathbf{L}\mathbf{z},
            \quad \mathbf{z} \sim \mathcal{N}(\mathbf{0}, \mathbf{I})

        where :math:`\mathbf{L}` is the lower Cholesky factor of the covariance.

        Parameters
        ----------
        key : KeyArray
            JAX PRNG key.
        sample_shape : tuple of int, optional
            Leading batch dimensions for the samples. Defaults to ``()``,
            returning a single sample.

        Returns
        -------
        Float[Array, "... N"]
            Array of samples with shape ``(*sample_shape, N)``.
        """
        assert is_prng_key(key)
        # Obtain covariance root.
        covariance_root = lower_cholesky(self.scale)

        # Gather n samples from standard normal distribution Z = [z₁, ..., zₙ]ᵀ.
        white_noise = jr.normal(
            key, shape=sample_shape + self.batch_shape + self.event_shape
        )

        # xᵢ ~ N(loc, cov) <=> xᵢ = loc + sqrt zᵢ, where zᵢ ~ N(0, I).
        def affine_transformation(_x):
            return self.loc + covariance_root @ _x

        if not sample_shape:
            return affine_transformation(white_noise)

        return vmap(affine_transformation)(white_noise)

    @property
    def mean(self) -> Float[Array, " N"]:
        r"""Calculates the mean."""
        return self.loc

    @property
    def variance(self) -> Float[Array, " N"]:
        r"""Calculates the marginal variance (diagonal of the covariance)."""
        return diag(self.scale)

    def entropy(self) -> ScalarFloat:
        r"""Calculates the differential entropy of the distribution.

        .. math::

            H[p] = \tfrac{1}{2}\bigl(N(1 + \ln 2\pi) + \ln|\mathbf{\Sigma}|\bigr)

        Returns
        -------
        ScalarFloat
            Entropy in nats.
        """
        return 0.5 * (
            self.event_shape[0] * (1.0 + jnp.log(2.0 * jnp.pi)) + logdet(self.scale)
        )

    def median(self) -> Float[Array, " N"]:
        r"""Calculates the median (equal to the mean for a Gaussian)."""
        return self.loc

    def mode(self) -> Float[Array, " N"]:
        r"""Calculates the mode (equal to the mean for a Gaussian)."""
        return self.loc

    def covariance(self) -> Float[Array, "N N"]:
        r"""Materialises the full covariance matrix as a dense array.

        Returns
        -------
        Float[Array, "N N"]
            Dense covariance matrix.
        """
        return self.scale.to_dense()

    @property
    def covariance_matrix(self) -> Float[Array, "N N"]:
        r"""Property alias for :meth:`covariance`."""
        return self.covariance()

    def stddev(self) -> Float[Array, " N"]:
        r"""Calculates the marginal standard deviation."""
        return jnp.sqrt(diag(self.scale))

    def log_prob(self, y: Float[Array, " N"]) -> ScalarFloat:
        r"""Calculates the log pdf of the multivariate Gaussian.

        .. math::

            \log p(\mathbf{y}) = -\tfrac{1}{2}\bigl[
                N\ln 2\pi + \ln|\mathbf{\Sigma}|
                + (\mathbf{y} - \boldsymbol{\mu})^\top
                  \mathbf{\Sigma}^{-1}
                  (\mathbf{y} - \boldsymbol{\mu})
            \bigr]

        Parameters
        ----------
        y : Float[Array, " N"]
            Point at which to evaluate the log-density.

        Returns
        -------
        ScalarFloat
            Log probability.
        """
        mu = self.loc
        sigma = self.scale
        n = mu.shape[-1]

        # diff, y - µ
        diff = y - mu

        # compute the pdf, -1/2[ n log(2π) + log|Σ| + (y - µ)ᵀΣ⁻¹(y - µ) ]
        return -0.5 * (
            n * jnp.log(2.0 * jnp.pi) + logdet(sigma) + diff.T @ solve(sigma, diff)
        )

    def kl_divergence(self, other: "GaussianDistribution") -> ScalarFloat:
        r"""KL divergence from ``self`` to ``other``.

        Computes :math:`\operatorname{KL}[q \| p]` where ``self`` is *q* and
        ``other`` is *p*.

        Parameters
        ----------
        other : GaussianDistribution
            The reference distribution *p*.

        Returns
        -------
        ScalarFloat
            KL divergence in nats.
        """
        return _kl_divergence(self, other)


def _check_and_return_dimension(
    q: GaussianDistribution, p: GaussianDistribution
) -> int:
    r"""Checks that the dimensions of the distributions are compatible."""
    if q.event_shape != p.event_shape:
        raise ValueError(
            "Distribution event shapes are not compatible: `q.event_shape ="
            f" {q.event_shape}` and `p.event_shape = {p.event_shape}`. Please check"
            " your mean and covariance shapes."
        )

    return q.event_shape[-1]


def _frobenius_norm_squared(matrix: Float[Array, "N N"]) -> ScalarFloat:
    r"""Calculates the squared Frobenius norm of a matrix."""
    return jnp.sum(jnp.square(matrix))


def _kl_divergence(q: GaussianDistribution, p: GaussianDistribution) -> ScalarFloat:
    r"""KL-divergence between two Gaussians.

    Computes the KL divergence, $\operatorname{KL}[q\mid\mid p]$, between two
    multivariate Gaussian distributions $q(x) = \mathcal{N}(x; \mu_q, \Sigma_q)$
    and $p(x) = \mathcal{N}(x; \mu_p, \Sigma_p)$.

    Args:
        q: a multivariate Gaussian distribution.
        p: another multivariate Gaussian distribution.

    Returns:
        ScalarFloat: The KL divergence between q and p.
    """
    n_dim = _check_and_return_dimension(q, p)

    # Extract q mean and covariance.
    mu_q = q.loc
    sigma_q = q.scale

    # Extract p mean and covariance.
    mu_p = p.loc
    sigma_p = p.scale

    # Find covariance roots.
    sqrt_p = lower_cholesky(sigma_p)
    sqrt_q = lower_cholesky(sigma_q)

    # diff, μp - μq
    diff = mu_p - mu_q

    # trace term, tr[Σp⁻¹ Σq] = tr[(LpLpᵀ)⁻¹(LqLqᵀ)] = tr[(Lp⁻¹Lq)(Lp⁻¹Lq)ᵀ] = (fr[LqLp⁻¹])²
    trace = _frobenius_norm_squared(
        solve(sqrt_p, sqrt_q.to_dense())
    )  # TODO: Not most efficient, given the `to_dense()` call (e.g., consider diagonal p and q). Need to abstract solving linear operator against another linear operator.

    # Mahalanobis term, (μp - μq)ᵀ Σp⁻¹ (μp - μq) = tr [(μp - μq)ᵀ [LpLpᵀ]⁻¹ (μp - μq)] = (fr[Lp⁻¹(μp - μq)])²
    mahalanobis = jnp.sum(jnp.square(solve(sqrt_p, diff)))

    # KL[q(x)||p(x)] = [ [(μp - μq)ᵀ Σp⁻¹ (μp - μq)] - n - log|Σq| + log|Σp| + tr[Σp⁻¹ Σq] ] / 2
    return (mahalanobis - n_dim - logdet(sigma_q) + logdet(sigma_p) + trace) / 2.0


__all__ = [
    "GaussianDistribution",
]



# Copyright 2024 The GPJax Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Tests that core GPJax operations produce correct dtypes under float32 and
float64.  The global ``jax_enable_x64`` flag is toggled per-test via the
*enable_x64* fixture so these tests exercise both precision modes.
"""

from gpjax.dataset import Dataset
from gpjax.gps import Prior
from gpjax.kernels.stationary import RBF, Matern32
from gpjax.likelihoods import Gaussian
from gpjax.mean_functions import Constant
from gpjax.objectives import conjugate_mll
from jax import config
import jax.numpy as jnp
import pytest

config.update("jax_enable_x64", True)

KERNELS = [RBF, Matern32]


@pytest.fixture(params=[jnp.float32, jnp.float64], ids=["float32", "float64"])
def enable_x64(request):
    """Toggle x64 mode and yield the target dtype."""
    dtype = request.param
    use_x64 = dtype == jnp.float64
    prev = config.jax_enable_x64
    config.update("jax_enable_x64", use_x64)
    try:
        yield dtype
    finally:
        config.update("jax_enable_x64", prev)


# --------------------------------------------------------------------------- #
#  Kernel dtype tests                                                          #
# --------------------------------------------------------------------------- #


@pytest.mark.parametrize("KernelClass", KERNELS)
def test_kernel_gram_dtype(KernelClass, enable_x64):
    dtype = enable_x64
    x = jnp.linspace(0, 1, 10, dtype=dtype)[:, None]
    kernel = KernelClass()
    Kxx = kernel.gram(x).to_dense()
    assert Kxx.dtype == dtype, f"Expected {dtype}, got {Kxx.dtype}"


@pytest.mark.parametrize("KernelClass", KERNELS)
def test_kernel_cross_covariance_dtype(KernelClass, enable_x64):
    dtype = enable_x64
    x = jnp.linspace(0, 1, 10, dtype=dtype)[:, None]
    y = jnp.linspace(0.5, 1.5, 5, dtype=dtype)[:, None]
    kernel = KernelClass()
    Kxy = kernel.cross_covariance(x, y)
    assert Kxy.dtype == dtype, f"Expected {dtype}, got {Kxy.dtype}"


# --------------------------------------------------------------------------- #
#  GP prior dtype tests                                                        #
# --------------------------------------------------------------------------- #


def test_prior_predict_dtype(enable_x64):
    dtype = enable_x64
    x = jnp.linspace(0, 1, 10, dtype=dtype)[:, None]
    prior = Prior(mean_function=Constant(), kernel=RBF())
    dist = prior.predict(x)
    assert dist.mean.dtype == dtype, (
        f"Mean dtype: expected {dtype}, got {dist.mean.dtype}"
    )


# --------------------------------------------------------------------------- #
#  GP posterior dtype tests                                                     #
# --------------------------------------------------------------------------- #


@pytest.mark.filterwarnings("ignore:X is not of type float64:UserWarning")
@pytest.mark.filterwarnings("ignore:y is not of type float64:UserWarning")
def test_posterior_predict_dtype(enable_x64):
    dtype = enable_x64
    x_train = jnp.linspace(0, 1, 20, dtype=dtype)[:, None]
    y_train = jnp.sin(x_train)
    D = Dataset(X=x_train, y=y_train)

    x_test = jnp.linspace(0, 1, 5, dtype=dtype)[:, None]
    prior = Prior(mean_function=Constant(), kernel=RBF())
    likelihood = Gaussian(num_datapoints=D.n)
    posterior = prior * likelihood

    dist = posterior.predict(x_test, D)
    assert dist.mean.dtype == dtype, (
        f"Mean dtype: expected {dtype}, got {dist.mean.dtype}"
    )


# --------------------------------------------------------------------------- #
#  Objective dtype tests                                                       #
# --------------------------------------------------------------------------- #


@pytest.mark.filterwarnings("ignore:X is not of type float64:UserWarning")
@pytest.mark.filterwarnings("ignore:y is not of type float64:UserWarning")
def test_conjugate_mll_dtype(enable_x64):
    dtype = enable_x64
    x = jnp.linspace(0, 1, 20, dtype=dtype)[:, None]
    y = jnp.sin(x)
    D = Dataset(X=x, y=y)

    prior = Prior(mean_function=Constant(), kernel=RBF())
    likelihood = Gaussian(num_datapoints=D.n)
    posterior = prior * likelihood

    mll = conjugate_mll(posterior, D)
    assert mll.dtype == dtype, f"MLL dtype: expected {dtype}, got {mll.dtype}"
